import asyncio
import time
import typing as t

import pytest

import imy.async_utils

T = t.TypeVar("T")


_count_concurrent_running = 0


async def _async_generator(n: int) -> t.AsyncIterable[int]:
    """
    Like `range`, but asynchronous.
    """
    for ii in range(n):
        yield ii


async def _count_concurrent(value: T) -> tuple[T, int]:
    """
    Returns the input value as is, as well as how often this function was
    running concurrently during the invocation. In order for this to work the
    function imposes a small delay.
    """
    global _count_concurrent_running
    _count_concurrent_running += 1

    try:
        await asyncio.sleep(0.1)
        return value, _count_concurrent_running
    finally:
        _count_concurrent_running -= 1


@pytest.mark.asyncio
async def test_collect_basic() -> None:
    result = await imy.async_utils.collect(_async_generator(5))
    assert result == [0, 1, 2, 3, 4]


@pytest.mark.asyncio
async def test_collect_limit_too_high() -> None:
    result = await imy.async_utils.collect(_async_generator(5), limit=10)
    assert result == [0, 1, 2, 3, 4]


@pytest.mark.asyncio
async def test_collect_limit_too_low() -> None:
    result = await imy.async_utils.collect(_async_generator(5), limit=3)
    assert result == [0, 1, 2]


@pytest.mark.asyncio
async def test_collect_empty() -> None:
    result = await imy.async_utils.collect(_async_generator(0))
    assert result == []


@pytest.mark.asyncio
async def test_amap_simple() -> None:
    results = [
        ii
        async for ii in imy.async_utils.amap(
            _count_concurrent,
            range(5),
            concurrency=100,
        )
    ]

    for ii, (value, concurrent) in enumerate(results):
        assert value == ii


@pytest.mark.asyncio
async def test_amap_concurrency() -> None:
    results = [
        ii
        async for ii in imy.async_utils.amap(
            _count_concurrent,
            range(5),
            concurrency=2,
        )
    ]

    for ii, (value, concurrent) in enumerate(results):
        assert value == ii
        assert concurrent <= 2


# TODO: Test that `amap` returns exceptions ASAP


@pytest.mark.asyncio
async def test_iterator_to_thread_simple() -> None:
    results = [
        val
        async for val in imy.async_utils.iterator_to_thread(
            range(5),
            batch_size=5,
        )
    ]

    for value_should, value_is in enumerate(results):
        assert value_should == value_is


# TODO: More tests for `iterator_to_thread`


@pytest.mark.asyncio
async def test_multitask_empty() -> None:
    """
    Calls `imy.async_utils.multitask` without any tasks. This should
    immediately.
    """
    start_time = time.monotonic()

    await imy.async_utils.multitask()

    end_time = time.monotonic()
    assert end_time - start_time < 0.1


@pytest.mark.asyncio
async def test_multitask_concurrency() -> None:
    """
    Runs several functions via `imy.async_utils.multitask` and ensures that they
    run concurrently.
    """

    async def _slow_function() -> None:
        await asyncio.sleep(0.1)

    start_time = time.monotonic()

    await imy.async_utils.multitask(
        _slow_function(),
        _slow_function(),
        _slow_function(),
    )

    end_time = time.monotonic()
    assert end_time - start_time < 0.2


@pytest.mark.asyncio
async def test_multitask_exception() -> None:
    """
    Tests that `imy.async_utils.multitask` cancels all futures if an exception
    occurs and that the exception is re-raised.
    """

    async def raising_function() -> None:
        await asyncio.sleep(0.1)
        raise ValueError("Test")

    # Start the raising function as well as a much slower one. The exception
    # should be raised, and the whole thing return much faster than if the
    # slow function had been allowed to finish.
    start_time = time.monotonic()

    with pytest.raises(ValueError):
        await imy.async_utils.multitask(
            raising_function(),
            asyncio.sleep(1),
        )

    end_time = time.monotonic()
    assert end_time - start_time < 0.2


@pytest.mark.asyncio
async def test_to_daemon_thread_with_args() -> None:
    """
    Make sure that `imy.async_utils.to_daemon_thread` returns the result of the
    called function and passes through parameters correctly.
    """

    def callback(x: int, y: int) -> int:
        return x + y

    result = await imy.async_utils.to_daemon_thread(callback, 1, 2)
    assert result == 3


@pytest.mark.asyncio
async def test_to_daemon_thread_with_kwargs() -> None:
    """
    Make sure that `imy.async_utils.to_daemon_thread` returns the result of the
    called function and passes through parameters correctly.
    """

    def sync_function(x: int, y: int, z: int = 0) -> int:
        return x + y + z

    result = await imy.async_utils.to_daemon_thread(sync_function, x=1, y=2, z=3)
    assert result == 6


@pytest.mark.asyncio
async def test_to_daemon_thread_with_args_and_kwargs() -> None:
    """
    Make sure that `imy.async_utils.to_daemon_thread` returns the result of the
    called function and passes through parameters correctly.
    """

    def sync_function(x: int, y: int, z: int = 0) -> int:
        return x + y + z

    result = await imy.async_utils.to_daemon_thread(sync_function, 1, 2, z=3)
    assert result == 6


@pytest.mark.asyncio
async def test_to_daemon_thread_exception() -> None:
    """
    Tests that `imy.async_utils.to_daemon_thread` propagates exceptions from the
    called function.
    """

    def sync_function(x: int, y: int) -> int:
        raise ValueError("Test exception")

    with pytest.raises(ValueError, match="Test exception"):
        await imy.async_utils.to_daemon_thread(sync_function, 1, 2)


@pytest.mark.asyncio
async def test_to_daemon_thread_concurrency() -> None:
    """
    Tests that `imy.async_utils.to_daemon_thread` doesn't block execution of
    the current task
    """

    def slow_function() -> None:
        time.sleep(0.3)

    start_time = time.monotonic()

    # Give imy something to do
    asyncio.create_task(imy.async_utils.to_daemon_thread(slow_function))

    # Make sure the other task gets some execution time
    await asyncio.sleep(0.1)

    # This piece of code should be reached before the slow function finishes
    end_time = time.monotonic()
    assert end_time - start_time < 0.2



"""
Utilities for setting up Python's weirdo logging system. Including logging to
MongoDB if you feel crazy.
"""

from __future__ import annotations

import atexit
import dataclasses
import logging
import logging.handlers
import queue
import socket
import threading
import time
import typing as t
import weakref
from datetime import datetime, timezone
from pathlib import Path

from . import async_utils

try:
    import pymongo.collection
    import uniserde
    from uniserde import BsonDoc, ObjectId
except ImportError:
    if t.TYPE_CHECKING:
        import pymongo.collection
        import uniserde
        from uniserde import BsonDoc, ObjectId
else:
    bson_serde = uniserde.BsonSerde.new_camel_case()


__all__ = [
    "LogLevel",
    "LOG_LEVEL_NAMES",
    "LOG_LEVEL_ANNOUNCEMENT",
    "Environment",
    "levels_greater_than_or_equal",
    "MongoDbLogger",
    "setup_logging",
]


T = t.TypeVar("T")


LogLevel: t.TypeAlias = t.Literal[
    "debug",
    "info",
    "announcement",
    "warning",
    "error",
    "critical",
]
LOG_LEVEL_NAMES = t.get_args(LogLevel)  # In ascending order!

# Add the announcement level to the logging module
LOG_LEVEL_ANNOUNCEMENT = (logging.INFO + logging.WARNING) // 2
logging.addLevelName(LOG_LEVEL_ANNOUNCEMENT, "ANNOUNCEMENT")

Environment: t.TypeAlias = t.Literal["development", "production"]


# Keep track of all living persistence loggers, so they can be flushed at exit
_ALL_MONGODB_LOGGERS: weakref.WeakSet[MongoDbLogger] = weakref.WeakSet()


@atexit.register
def _flush_all_persistence_loggers() -> None:
    for logger in _ALL_MONGODB_LOGGERS:
        logger.flush_sync()


def _log_level_to_python(level: LogLevel) -> int:
    return logging.getLevelName(level.upper())


def _log_level_from_python(level: int) -> LogLevel:
    result = logging.getLevelName(level).lower()
    assert result in LOG_LEVEL_NAMES, f"Unknown log level: {result}"
    return result  # type: ignore


def levels_greater_than_or_equal(level: LogLevel) -> set[LogLevel]:
    """
    Returns a set containing all log levels which are at least as severe as the
    given one.

    ## Raises

    `ValueError`: if the given level is not a valid log level.
    """
    for ii, name in enumerate(LOG_LEVEL_NAMES):
        if name == level:
            return set(LOG_LEVEL_NAMES[ii:])

    raise ValueError(f"Unknown log level: {level}")


@dataclasses.dataclass
class LogEntry:
    """
    Represents a single log entry in the database.
    """

    # Unique identifier for the log entry
    id: ObjectId

    # When this log entry was created
    timestamp: datetime

    # Whether this log entry was created in development or production
    environment: Environment

    # Which machine created this log entry
    host: str

    # The app that created this log entry
    app: str

    # The severity of this log entry
    level: LogLevel

    # A human-readable message. This is the main content of the log entry
    message: str

    # Arbitrary additional data
    payload: dict[str, t.Any]


class MongoDbLogger(logging.StreamHandler):
    """
    Logger, which stores its entries in a MongoDB database.

    Database access is asynchronous, and likely on another server. `await`ing
    every log operation would be incredibly slow. Instead, this logger only
    synchronously queues log entries, and then later asynchronously copies them
    to the database.
    """

    def __init__(
        self,
        collection: pymongo.collection.Collection,
        *,
        environment: Environment,
        app: str,
        host: str | None = None,
    ):
        super().__init__()

        self._sync_collection = collection

        self._host = socket.gethostname() if host is None else host
        self._environment: Environment = environment
        self._app = app

        self._last_log_writeback_time_monotonic = time.monotonic()
        self._pending_log_entries: queue.Queue[LogEntry] = queue.Queue()

        self._writeback_thread = threading.Thread(
            target=self._writeback_worker, daemon=True
        )
        self._writeback_thread.start()

        # Log entries are plentiful. Make sure there's an index
        collection.create_index([("timestamp", 1)])

        # Keep track of all living persistence loggers
        _ALL_MONGODB_LOGGERS.add(self)

    def __del__(self) -> None:
        self.flush_sync()

    def create_log_entries_sync(self, entries: t.Iterable[LogEntry]) -> None:
        """
        Batch creates new log entries in the database.

        The entries need to have unique ids among all entries in the database.
        This is not checked for performance reasons.
        """
        # MongoDB doesn't like empty inserts
        entry_data = [bson_serde.as_bson(entry) for entry in entries]

        if not entry_data:
            return

        # Insert the entries
        self._sync_collection.insert_many(entry_data)

    def flush_sync(self) -> None:
        """
        Copies any not yet stored log entries into the database.
        """
        # Move the log entries into a local variable
        in_flight_entries = []

        while True:
            try:
                in_flight_entries.append(self._pending_log_entries.get_nowait())
            except queue.Empty:
                break

        # Try to push the entries into the database
        try:
            self.create_log_entries_sync(in_flight_entries)

        # If the operation fails, put the entries back into the queue so they
        # can be retried later
        except Exception:
            for entry in in_flight_entries:
                self._pending_log_entries.put(entry)

            raise

    def _writeback_worker(self) -> None:
        # Keep copying entries forever. There is no need to return because the
        # thread is daemonized.
        #
        # TODO: What if the logger is deleted? This isn't currently supported.
        CYCLE_TIME = 10

        while True:
            # Wait some time before writing back. This batches log entries
            # together.
            now = time.monotonic()
            sleep_time = CYCLE_TIME - (now - self._last_log_writeback_time_monotonic)

            if sleep_time > 0:
                time.sleep(sleep_time)

            # Copy the pending log entries into the database
            try:
                self.flush_sync()
            except Exception as e:
                logging.error(f"Error writing back log entries: {e}")

                # Wait a bit before retrying
                time.sleep(20)

            # Housekeeping
            self._last_log_writeback_time_monotonic = time.monotonic()

    def queue_log(
        self,
        level: LogLevel,
        message: str,
        *,
        payload: dict[str, t.Any] | None = None,
    ) -> None:
        """
        Creates a log entry and queues it for storage to the database. The entry
        will be created later, or once `flush` is called.
        """
        self._pending_log_entries.put(
            LogEntry(
                id=ObjectId(),
                timestamp=datetime.now(timezone.utc),
                environment=self._environment,
                host=self._host,
                app=self._app,
                level=level,
                message=message,
                payload={} if payload is None else payload,
            )
        )

    def emit(self, record: logging.LogRecord) -> None:
        """
        For compatibility with python's logging module.
        """

        self.queue_log(
            level=_log_level_from_python(record.levelno),
            message=record.message,
        )

    def find_log_entries(
        self,
        *,
        environment: Environment | None = None,
        levels: t.Iterable[LogLevel] | None = None,
        newer_than: datetime | None = None,
        older_than: datetime | None = None,
        limit: int | None = None,
    ) -> t.AsyncIterable[LogEntry]:
        """
        Returns an async iterator over all log entries in the database matching
        the given filters.

        The entries are sorted by timestamp, with the most recent entries first.
        """
        # Build the query
        query: BsonDoc = {}

        if environment is not None:
            query["environment"] = bson_serde.as_bson(environment)

        if levels is not None:
            query["level"] = {"$in": [bson_serde.as_bson(level) for level in levels]}

        if newer_than is not None:
            query["timestamp"] = {"$gt": newer_than}

        if older_than is not None:
            query["timestamp"] = {"$lt": older_than}

        # Yield all matches
        def sync_iterator() -> t.Iterator[LogEntry]:
            cursor = self._sync_collection.find(query).sort([("timestamp", -1)])

            if limit is not None:
                cursor = cursor.limit(limit)

            for doc in cursor:
                yield bson_serde.from_bson(LogEntry, doc)

        return async_utils.iterator_to_thread(sync_iterator(), batch_size=50)

    def watch_log_entries(
        self,
        *,
        environment: Environment | None = None,
        levels: t.Iterable[LogLevel] | None = None,
    ) -> t.AsyncIterable[LogEntry]:
        """
        Returns an async iterator over all new log entries in the database
        matching the given filters.

        The iterator will block until new log entries are available.
        """
        # Build the filter pipeline
        query: BsonDoc = {"operationType": "insert"}

        if environment is not None:
            query["fullDocument.environment"] = bson_serde.as_bson(environment)

        if levels is not None:
            query["fullDocument.level"] = {
                "$in": [bson_serde.as_bson(level) for level in levels]
            }

        pipeline = [
            {
                "$match": query,
            }
        ]

        # Watch for changes
        def sync_iterator() -> t.Iterator[LogEntry]:
            for change in self._sync_collection.watch(pipeline=pipeline):
                yield bson_serde.from_bson(
                    LogEntry,
                    change["fullDocument"],
                )

        return async_utils.iterator_to_thread(sync_iterator(), batch_size=1)


def reset_logging() -> None:
    """
    Resets the logging system to its initial state. This removes all handlers
    from the root logger and resets its level to `WARNING`.
    """

    # Get the root logger
    root_logger = logging.getLogger()

    # Remove all existing handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Reset the root logger level
    root_logger.setLevel(logging.WARNING)


@t.overload
def setup_logging(
    *,
    info_log_path: Path | None = None,
    debug_log_path: Path | None = None,
    stdout_log_level: LogLevel = "info",
) -> MongoDbLogger: ...


@t.overload
def setup_logging(
    *,
    info_log_path: Path | None = None,
    debug_log_path: Path | None = None,
    database_collection: pymongo.collection.Collection,
    database_app: str,
    database_environment: Environment,
    database_host: str | None = None,
    stdout_log_level: LogLevel = "info",
    database_log_level: LogLevel = "debug",
) -> MongoDbLogger: ...


def setup_logging(
    *,
    info_log_path: Path | None = None,
    debug_log_path: Path | None = None,
    database_collection: pymongo.collection.Collection | None = None,
    database_app: str | None = None,
    database_environment: Environment | None = None,
    database_host: str | None = None,
    stdout_log_level: LogLevel = "info",
    database_log_level: LogLevel = "debug",
) -> MongoDbLogger | None:
    """
    Creates a nice logging setup. Any previously registered handlers are
    removed, which means it is safe to call this function multiple times without
    creating duplicate log entries.

    - INFO logs to `info_log_path`, keeping logs indefinitely
    - Configurable logs to `stdout`, INFO by default
    - DEBUG logs to `debug_log_path`, keeping a limited number of days
    - DEBUG logs into the database, if a collection is provided

    Returns the database logger, if one was created.

    ## Parameters

    `info_log_path`: Path to the info log file

    `debug_log_path`: Path to the debug log file

    `database_collection`: The MongoDB collection to store logs in

    `database_app`: The running app's name. This will be added to logged
        entries in the database

    `database_environment`: The environment that will appear in the database
        log entries. This will allow you to distinguish between development and
        production logs.

    `database_host`: Override for the machine's name that will appear in the
        database log entries. If not provided, the machine's hostname will be
        used.

    `stdout_log_level`: The minimum log level to display on stdout

    `database_log_level`: The minimum log level to store in the database
    """

    # Make sure not to add multiple handlers if the function is called multiple
    # times.
    reset_logging()

    # Configure the root logger and prepare some values
    root_logger = logging.getLogger("")
    root_logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter("%(asctime)s  %(levelname)-8s  %(message)s")

    # Info -> file
    if info_log_path is not None:
        info_log_path.parent.mkdir(parents=True, exist_ok=True)

        handler = logging.handlers.TimedRotatingFileHandler(
            info_log_path,
            encoding="utf-8",
            when="midnight",
            utc=True,
        )
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    # Stdout (configurable)
    handler = logging.StreamHandler()
    handler.setLevel(_log_level_to_python(stdout_log_level))
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    # Debug -> file
    if debug_log_path is not None:
        debug_log_path.parent.mkdir(parents=True, exist_ok=True)

        handler = logging.handlers.TimedRotatingFileHandler(
            debug_log_path,
            encoding="utf-8",
            when="midnight",
            utc=True,
            backupCount=7,
        )
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    # Database
    if database_collection is None:
        pers_logger = None
    else:
        assert (
            database_environment is not None
        ), "Must provide an environment when logging to a database"

        assert (
            database_app is not None
        ), "Must provide an app name when logging to a database"

        pers_logger = MongoDbLogger(
            database_collection,
            host=database_host,
            environment=database_environment,
            app=database_app,
        )

        pers_logger.setLevel(_log_level_to_python(database_log_level))
        root_logger.addHandler(pers_logger)

    return pers_logger



"""
A very simple dependency injection system, in pure Python.

To use this, create an `Injector` instance, and bind factory functions to it.
Each factory may depend on types provided by other factories, and returns a
value, which is then stored in the injector.

When a value is requested from the injector, it will either return an existing
value if one is already available, or call the appropriate factory function(s)
to create it.

For convenience, a global injector is also provided, which can be accessed using
global functions such as `get` and `bind`.
"""

from __future__ import annotations

import inspect
import io
import typing as t

import introspection.typing
import typing_extensions as te

__all__ = [
    "DependencyCycleError",
    "Injector",
    "get",
    "set",
    "bind",
]

T = t.TypeVar("T")
P = te.ParamSpec("P")
R = t.TypeVar("R")


def _key_from_type_annotation(
    annotation: t.Any,
    forward_ref_context: str,
) -> t.Type:
    """
    Given a type annotation, return the type that should be used as a key.
    """
    info = introspection.typing.TypeInfo(
        annotation,
        forward_ref_context=forward_ref_context,
    )

    if info.type is None:
        raise ValueError("`None` cannot be used as a key.")

    assert inspect.isclass(info.type), (annotation, info.type)

    return info.type  # type: ignore


def _key_from_type(parameter: t.Type) -> t.Type:
    """
    Given a type as would be passed into a function (like __getitem__), return
    the type that should be used as a key.

    This is in contrast to a type annotation, which may contain strings, forward
    references, etc.
    """
    origin = t.get_origin(parameter)
    key = parameter if origin is None else parameter

    assert inspect.isclass(key), (parameter, key)

    return key


class DependencyCycleError(Exception):
    def __init__(self, sequence: t.Sequence[t.Type]) -> None:
        super().__init__()
        self._sequence = sequence

    def __str__(self) -> str:
        f = io.StringIO()

        f.write(f"<{self.__class__.__name__} ")
        f.write(" -> ".join(key.__name__ for key in self._sequence))
        f.write(">")

        return f.getvalue()


def _parse_factory(factory: t.Callable) -> tuple[list[t.Type], t.Type]:
    """
    Given a factory function, return a tuple containing the set of types it
    depends on and the type of item generated. The types are extracted from the
    factory's signature and order is preserved.
    """
    signature = inspect.signature(factory)

    # TODO: What if the factory uses kwargs?

    # Parameters
    parameter_types: list[t.Type] = []

    for parameter in signature.parameters.values():
        # Make sure the parameter has a type annotation
        if parameter.annotation is parameter.empty:
            raise ValueError(
                f"All parameters of factory functions need to have type annotations. `{parameter.name}` is missing one."
            )

        # Return type
        parameter_types.append(
            _key_from_type_annotation(
                parameter.annotation,
                factory.__module__,
            )
        )

    # Return type
    return_type = signature.return_annotation

    if return_type is signature.empty:
        raise ValueError("Factory functions need to have a return type annotation.")

    return_type_key = _key_from_type_annotation(
        return_type,
        factory.__module__,
    )

    # Done
    return parameter_types, return_type_key


class Injector:
    """
    TODO

    Note on generic keys (e.g. "list[int]"): Factory functions may use arbitrarily
    complex type annotations, however only the base type is used as a key. This
    means that even though a `list[int]` may be attached, during retrieval,
    `list[str]` will return the same, in this case incorrect, instance.

    This is a compromise to keep the system simple and fast, since hashing
    recursive types would be slow and is rarely useful anyway.
    """

    def __init__(
        self,
        *,
        items: t.Iterable[t.Any] = [],
    ) -> None:
        # Keeps track of all components currently attached to the injector
        self._components: dict[t.Type, t.Any] = {type(item): item for item in items}

        # Factories can be used to create items if they are not already
        # available in the injector. Each factory may depend on any number of
        # other items and returns the newly created item.
        self._factories: dict[t.Type, tuple[list[t.Type], t.Callable]] = {}

    def _get(
        self,
        raw_key: t.Type[T],
        *,
        in_flight: list[t.Type],
    ) -> T:
        """
        Helper function for `__getitem__`.
        """
        # Get the actual key to use
        key = _key_from_type(raw_key)

        # If the item is already available in this injector, just return it
        try:
            return self._components[key]
        except KeyError:
            pass

        # If an item of this type is currently being constructed, there's a
        # dependency cycle. Report it.
        if key in in_flight:
            raise DependencyCycleError(in_flight + [key])

        # Try to find a factory that can create the item
        try:
            factory_param_types, factory = self._factories[key]
        except KeyError:
            raise KeyError(raw_key)

        # Get all of the factory's dependencies
        factory_params: list[t.Any] = []

        for param_type in factory_param_types:
            factory_params.append(
                self._get(
                    param_type,
                    in_flight=in_flight + [key],
                )
            )

        # Create the item, register it and return
        item = factory(*factory_params)
        self._components[key] = item

        return item

    def __getitem__(self, key: t.Type[T]) -> T:
        """
        Given an item type, return the component of that type.

        ## Raises

        `KeyError`: If no item of the given type is available in this injector.

        `DependencyCycleError`: If the item cannot be constructed due to a
            dependency cycle.
        """
        # Delegate to the helper function
        return self._get(
            key,
            in_flight=[],
        )

    def __setitem__(self, item_type: t.Type[T], item: T) -> None:
        """
        Adds the given item to the injector. If the item is already present, it
        will be replaced.

        ## Raises

        `ValueError`: If the item is not an instance of the given type.
        """
        # Get the actual key to use
        key = _key_from_type(item_type)

        # Make sure the item is of the correct type
        if not isinstance(item, key):
            raise ValueError(
                f"Item has to be of type `{item_type}`, not `{type(item)}`"
            )

        # Add the item to the injector
        self._components[key] = item

    def bind(
        self,
        as_type: t.Type | None = None,
    ) -> t.Callable[[t.Callable[P, R]], t.Callable[P, R]]:
        """
        Adds the decorated function to the injector as a factory. The function's
        result must be in instance of the given type or subclass thereof.

        ## Raises

        `ValueError`: If the function's return type is not a subclass of the given
            type.

        `TypeError`: If there is already a factory for the type.
        """

        def decorator(func: t.Callable[P, R]) -> t.Callable[P, R]:
            # Parse the factory
            param_keys, return_key = _parse_factory(func)

            if as_type is None:
                key_type = return_key
            else:
                key_type = _key_from_type(as_type)

                # Make sure the factory indeed returns the correct type
                if not issubclass(return_key, key_type):
                    raise ValueError(
                        f"The factory has to return values of type `{key_type}` (or any subclass), not `{return_key}`"
                    )

            # Make sure there isn't already a factory for this type
            if key_type in self._factories:
                raise TypeError(
                    f"There is already a factory for type `{key_type}`. Only one factory per type is allowed."
                )

            # Register the factory
            self._factories[key_type] = (param_keys, func)
            return func

        return decorator

    def clear(self) -> None:
        """
        Removes any previously added components and factories from the injector.
        """
        self._components.clear()
        self._factories.clear()


# Expose a global injector
GLOBAL_INJECTOR = Injector()

get = GLOBAL_INJECTOR.__getitem__
set = GLOBAL_INJECTOR.__setitem__
bind = GLOBAL_INJECTOR.bind
clear = GLOBAL_INJECTOR.clear



import imy.deprecations
import imy.docstrings


def _documented_function_names(docs: imy.docstrings.ClassDocs) -> set[str]:
    return set(docs.functions)


def _documented_attribute_names(docs: imy.docstrings.ClassDocs) -> set[str]:
    return set(docs.attributes)


class Parent:
    """
    # Leading Headings 1 are stripped

    This is the summary.

    This is the details. They can be very long and span multiple lines. They can
    even contain multiple paragraphs.

    Just like this one.

    ## Heading 2

    Any non-key sections are also part of the details.

    This is the end of the details.

    ## Attributes

    int_attribute: <int>

    `float_attribute`: <float>

    str_attribute: <str>

    ## Metadata

    public: False
    """

    int_attribute: int
    float_attribute: float
    str_attribute: str

    def __init__(self, foo, bar, baz):
        pass

    @property
    def bytes_property(self) -> bytes:
        """
        This is the getter function of `bytes_property`
        """
        return b""

    @bytes_property.setter
    def bytes_property(self, value: str | bytes) -> None:
        """
        This is the setter function of `bytes_property`
        """
        _ = value

    def numeric_function(self, x: int) -> float:
        """
        <function summary>

        <function details>

        ## Parameters

        x: <int>

        ## Raises

        `ValueError`: <raise-value-error>
        """
        return float(x)

    def overridden_function(self) -> float:
        """
        docstring for `overridden_function`
        """
        return 1.5

    def _private_because_of_underscore(self):
        pass

    def private_because_of_metadata(self):
        """
        ## Metadata

        `public`: False
        """

    def __str__(self) -> str:
        return "<Parent object>"


class Child(Parent):
    """
    Children are parents too!

    ## Attributes

    bool_attribute: <bool>

    ## Metadata

    public: True

    experimental: True
    """

    bool_attribute: bool

    async def list_function(self, x: str) -> list:
        """
        <function summary>

        <function details>

        ## Parameters

        x: <str>
        """
        return [x]

    def overridden_function(self) -> int:
        return 1


def test_parse_class_docstring() -> None:
    docs = imy.docstrings.ClassDocs.from_class(Parent)

    assert docs.name == "Parent"
    assert docs.summary == "This is the summary."

    assert docs.details is not None
    assert docs.details.startswith("This is the details.")
    assert docs.details.endswith("This is the end of the details.")

    assert _documented_function_names(docs) == {
        "__init__",
        "numeric_function",
        "overridden_function",
    }

    init_function_docs = docs.functions["__init__"]
    assert init_function_docs.owner is docs
    assert init_function_docs.name == "__init__"
    assert init_function_docs.synchronous is True
    assert init_function_docs.return_type is imy.docstrings.UNSET
    assert init_function_docs.summary is None
    assert init_function_docs.details is None

    numeric_function_docs = docs.functions["numeric_function"]
    assert numeric_function_docs.owner is docs
    assert numeric_function_docs.name == "numeric_function"
    assert numeric_function_docs.synchronous is True
    assert numeric_function_docs.return_type is float
    assert numeric_function_docs.summary == "<function summary>"
    assert numeric_function_docs.details == "<function details>"

    assert len(numeric_function_docs.parameters) == 2
    param1, param2 = numeric_function_docs.parameters.values()

    assert param1.owner is numeric_function_docs
    assert param1.name == "self"

    assert param2.owner is numeric_function_docs
    assert param2.name == "x"
    assert param2.type is int
    assert param2.description == "<int>"

    overridden_function_docs = docs.functions["overridden_function"]
    assert overridden_function_docs.owner is docs
    assert overridden_function_docs.name == "overridden_function"
    assert overridden_function_docs.synchronous is True
    assert overridden_function_docs.return_type is float
    assert overridden_function_docs.summary == "docstring for `overridden_function`"
    assert overridden_function_docs.details is None

    assert len(numeric_function_docs.raises) == 1
    assert numeric_function_docs.raises[0] == (
        "ValueError",
        "<raise-value-error>",
    )

    assert _documented_attribute_names(docs) == {
        "int_attribute",
        "float_attribute",
        "str_attribute",
    }

    for attr in docs.attributes.values():
        assert attr.owner is docs
        assert attr.type is not None
        assert attr.description is not None
        assert attr.description.strip() == f"<{attr.type.__name__}>"  # type: ignore

    for prop in docs.properties.values():
        assert prop.owner is docs

        assert prop.getter.owner is prop

        assert prop.setter is not None
        assert prop.setter.owner is prop

    assert docs.metadata.public is False
    assert docs.metadata.experimental is False


def test_parse_class_docstring_with_inheritance() -> None:
    docs = imy.docstrings.ClassDocs.from_class(Child)

    assert docs.name == "Child"
    assert docs.summary == "Children are parents too!"
    assert docs.details is None

    assert _documented_function_names(docs) == {
        "__init__",
        "numeric_function",
        "list_function",
        "overridden_function",
    }

    for func in docs.functions.values():
        assert func.owner is docs

        if func.name == "__init__":
            assert func.synchronous is True
            assert func.return_type is imy.docstrings.UNSET
            assert func.summary is None
            assert func.details is None

            assert len(func.parameters) == 4
            param1, param2, param3, param4 = func.parameters.values()

            assert param1.name == "self"

            assert param2.name == "foo"
            assert param2.type is imy.docstrings.UNSET
            assert param2.description is None

            assert param3.name == "bar"
            assert param3.type is imy.docstrings.UNSET
            assert param3.description is None

            assert param4.name == "baz"
            assert param4.type is imy.docstrings.UNSET
            assert param4.description is None

        elif func.name == "numeric_function":
            assert func.synchronous is True
            assert func.return_type is float
            assert func.summary == "<function summary>"
            assert func.details == "<function details>"

            assert len(func.parameters) == 2
            param1, param2 = func.parameters.values()

            assert param1.name == "self"

            assert param2.name == "x"
            assert param2.type is int
            assert param2.description == "<int>"

            assert len(func.raises) == 1
            assert func.raises[0] == ("ValueError", "<raise-value-error>")

        elif func.name == "list_function":
            assert func.synchronous is False
            assert func.return_type is list
            assert func.summary == "<function summary>"
            assert func.details == "<function details>"

            assert len(func.parameters) == 2
            param1, param2 = func.parameters.values()

            assert param1.name == "self"

            assert param2.name == "x"
            assert param2.type is str
            assert param2.description == "<str>"

            assert len(func.raises) == 0

        elif func.name == "overridden_function":
            assert func.owner is docs
            assert func.synchronous is True
            assert func.return_type is int
            assert func.summary == "docstring for `overridden_function`"
            assert func.details is None

        else:
            assert False, f"Unexpected function: {func.name}"

    assert _documented_attribute_names(docs) == {
        "int_attribute",
        "float_attribute",
        "str_attribute",
        "bool_attribute",
    }

    for attr in docs.attributes.values():
        assert attr.owner is docs
        assert attr.type is not None
        assert attr.description is not None
        assert attr.description.strip() == f"<{attr.type.__name__}>"  # type: ignore

    assert docs.metadata.public is True
    assert docs.metadata.experimental is True


def test_parse_module() -> None:
    from test import docstrings_test_package

    module_docs = imy.docstrings.ModuleDocs.from_module(docstrings_test_package)

    assert module_docs.object is docstrings_test_package
    assert module_docs.name == "docstrings_test_package"
    assert module_docs.owner is None
    assert module_docs.summary == "<docstrings_test_package summary>"
    assert module_docs.details == "<docstrings_test_package details>"
    assert module_docs.deprecations == []

    assert set(module_docs.members) == {"foo", "MyEnum"}

    foo = module_docs.members["foo"]
    assert isinstance(foo, imy.docstrings.FunctionDocs)
    assert foo.owner is module_docs

    MyEnum = module_docs.members["MyEnum"]
    assert isinstance(MyEnum, imy.docstrings.ClassDocs)
    assert MyEnum.owner is module_docs
    assert set(MyEnum.members) == {
        "is_even"
    }  # The __init__ should be omitted, since it's an enum

    module_docs.add_member(docstrings_test_package.public_submodule)
    assert set(module_docs.members) == {"foo", "MyEnum", "public_submodule"}

    public_submodule = module_docs.members["public_submodule"]
    assert isinstance(public_submodule, imy.docstrings.ModuleDocs)
    assert public_submodule.object is docstrings_test_package.public_submodule
    assert public_submodule.owner is module_docs
    assert set(public_submodule.members) == {"DemoClass"}


def test_deprecated_class():
    @imy.deprecations.deprecated(since="1.1", replacement="Bar")
    class Foo:
        pass

    docs = imy.docstrings.ClassDocs.from_class(Foo)
    assert len(docs.deprecations) == 1
    assert docs.deprecations[0].since_version == "1.1"
    assert docs.deprecations[0].will_be_removed_in_version is None
    assert "Bar" in docs.deprecations[0].message


def test_deprecated_function():
    @imy.deprecations.deprecated(
        since="1.2",
        will_be_removed_in_version="1.5",
        description="It causes undefined behavior",
    )
    @imy.deprecations.parameter_renamed(
        since="1.3", old_name="old_param", new_name="param"
    )
    async def foo(param: int):
        return

    docs = imy.docstrings.FunctionDocs.from_function(foo)

    assert len(docs.deprecations) == 1
    assert docs.deprecations[0].since_version == "1.2"
    assert docs.deprecations[0].will_be_removed_in_version == "1.5"
    assert docs.deprecations[0].message == "It causes undefined behavior"
    assert not docs.synchronous



"""
Re-implementation of task groups so they can be used in Python versions before
3.11

If the built-in `asyncio.TaskGroup` is available, it will be used, otherwise a
custom implementation is used to fill the gap.
"""

from __future__ import annotations

import asyncio
import typing as t


class BaseExceptionGroup(Exception):
    """
    Base class for exceptions that contain multiple exceptions.
    """

    def __init__(
        self,
        message: str,
        exceptions: t.Iterable[BaseException],
    ) -> None:
        """
        Initialize the exception with a message and a list of exceptions.
        """
        super().__init__(message)
        self.exceptions = list(exceptions)


class ExceptionGroup(BaseExceptionGroup):
    """
    An exception that contains multiple exceptions.
    """

    pass


class TaskGroup:
    """
    A context manager that manages a group of tasks.

    This is a re-implementation of `asyncio.TaskGroup` for Python versions
    before 3.11. See [the Python
    docs](https://docs.python.org/3/library/asyncio-task.html#task-groups)
    for how to use it.
    """

    def __init__(self) -> None:
        """
        Initialize the TaskGroup with default values.
        """
        # Set of tasks managed by the TaskGroup
        self._tasks: set[asyncio.Task] = set()

        # List of exceptions raised by the tasks
        self._exceptions: list[BaseException] = []

        # Flag indicating if the TaskGroup has been entered
        self._entered: bool = False

        # Flag indicating if the TaskGroup is exiting
        self._exiting: bool = False

    async def __aenter__(self) -> TaskGroup:
        """
        Enter the runtime context related to this object.
        """
        self._entered = True
        return self

    async def __aexit__(
        self,
        exc_type: t.Type[BaseException],
        exc_val: BaseException,
        exc_tb: t.Any,
    ) -> bool:
        """
        Exit the runtime context related to this object.
        """
        self._exiting = True

        # Cancel remaining tasks *before* waiting for them
        for task in self._tasks:
            if not task.done():
                task.cancel()

        # Wait for all tasks to complete, gathering exceptions
        await asyncio.gather(*self._tasks, return_exceptions=True)

        # Process the results of the tasks
        for task in self._tasks:
            if task.done():
                # Collect exceptions raised by the tasks
                try:
                    task.result()

                # CancelledError is expected and should be ignored
                except asyncio.CancelledError:
                    pass

                # You're coming with me!
                except Exception as e:
                    self._exceptions.append(e)

        # Handle exceptions raised during the context
        if exc_type is not None and exc_type is not asyncio.CancelledError:
            self._exceptions.append(exc_val)

        # Raise exceptions if any were collected
        if self._exceptions:
            # Some base exceptions are extra special
            for e in self._exceptions:
                if isinstance(e, (KeyboardInterrupt, SystemExit)):
                    raise e

            # Raise an ExceptionGroup or BaseExceptionGroup
            if all(isinstance(e, Exception) for e in self._exceptions):
                raise ExceptionGroup(
                    "TaskGroup raised exceptions",
                    self._exceptions,
                )
            else:
                raise BaseExceptionGroup(
                    "TaskGroup raised exceptions",
                    self._exceptions,
                )
        return False

    def create_task(
        self,
        coro: t.Coroutine,
        *,
        name: t.Optional[str] = None,
    ) -> asyncio.Task | None:
        """
        Creates a new task and runs it in the `TaskGroup`.
        """
        # If shutting down, no more tasks can be added
        if self._exiting:
            raise RuntimeError("This TaskGroup is shutting down")

        # Create and schedule the task
        task = asyncio.create_task(coro, name=name)
        self._tasks.add(task)

        # Remove task from the set when done
        task.add_done_callback(self._tasks.discard)

        # Done!
        return task



"""
<docstrings_test_package summary>

<docstrings_test_package details>
"""

from collections import defaultdict as defaultdict

from . import public_submodule as public_submodule
from .internal_module import *


def helper_function():
    """
    ## Metadata

    `public`: False
    """



from __future__ import annotations

import collections
import dataclasses
import enum
import inspect
import textwrap
import types
import typing as t
import warnings
from dataclasses import is_dataclass

import introspection
import introspection.types
import introspection.typing
from uniserde import Jsonable

from ..deprecations import deprecated_objects, deprecated_parameters
from . import data_models


def split_docstring_into_sections(docstring: str) -> tuple[str, dict[str, str]]:
    """
    Splits the given docstring into sections separated by markdown headings. The
    result is the part of the string before the first heading, and a dictionary
    mapping the heading names to the text of the sections ("the summary").

    If the docstring starts with a heading it will not create a separate section
    and the text of that section will be considered part of the summary.
    """
    # Drop the title if it exists
    if docstring.startswith("#") and not docstring.startswith("##"):
        docstring = docstring.split("\n", 1)[1]

    # Split the docstring into sections
    sections: dict[str, list[str]] = {}
    details: list[str] = []
    current_section: list[str] = details
    currently_inside_code_block: bool = False

    # Process the individual lines
    for line in docstring.splitlines():
        # Code block?
        if line.startswith("```"):
            currently_inside_code_block = not currently_inside_code_block
            current_section.append(line)

        # Section Header
        elif line.startswith("#") and not currently_inside_code_block:
            section_name = line.strip()
            current_section = sections.setdefault(section_name, [])

        # Nothing to see here
        else:
            current_section.append(line)

    # Post-process the sections
    def postprocess(section: list[str]) -> str:
        return "\n".join(section).strip()

    return postprocess(details), {
        name: postprocess(section) for name, section in sections.items()
    }


def parse_key_value_section(section_string: str) -> dict[str, str]:
    """
    Some docstring sections are formatted as a list of key-value pairs, such as
    this:

    ```
    key: A description of the key.
    `key_in_code`: A description of the key in code.
    ```

    (Note the optional backticks around the key - these are stripped if present)

    This function splits such a section into a dictionary of key-value pairs.
    """
    result_lines: dict[str, list[str]] = {}
    current_value: list[str] = []

    # Process the lines individually
    for raw_line in section_string.splitlines():
        # Strip the line and calculate the indentation
        strip_line = raw_line.lstrip()
        indent = len(raw_line) - len(strip_line)
        strip_line = strip_line.rstrip()

        # Continuation of the previous value
        if indent > 0 or not strip_line:
            current_value.append(raw_line)
            continue

        # New value?
        parts = strip_line.split(":", 1)

        # Kinda, but it's mistyped
        if len(parts) == 1:
            # TODO: Warn somehow?
            current_value.append(raw_line)
            continue

        # Yes, for real this time
        key, value = parts
        key = key.strip().strip("`").strip()
        value = value.lstrip()
        current_value = [value]
        result_lines[key] = current_value

    # Postprocess the values. The lines need to be joined and the values
    # dedented.
    result: dict[str, str] = {}

    for key, lines in result_lines.items():
        assert lines, (key, lines)

        # The first line has weird indendation, due to following the separator.
        # Strip it.
        value = lines[0].lstrip()

        # The remaining lines need to be dedented
        if len(lines) > 1:
            tail = "\n".join(lines[1:])
            tail = textwrap.dedent(tail)
            value = f"{value}\n{tail}"

        result[key] = value

    return result


def parse_details(details: str) -> tuple[str | None, str | None]:
    """
    Given the details part of a docstring, split it into summary and details.
    Either value may be Nonne if they are not present in the original string.
    """
    details = details.strip()

    # Split into summary & details
    lines = details.split("\n")

    short_lines: list[str] = []
    long_lines: list[str] = []
    cur_lines = short_lines

    for raw_line in lines:
        strip_line = raw_line.strip()

        cur_lines.append(raw_line)

        if not strip_line:
            cur_lines = long_lines

    # Join the lines
    short_description = "\n".join(short_lines).strip()
    long_description = "\n".join(long_lines).strip()

    if not short_description:
        short_description = None

    if not long_description:
        long_description = None

    # Done
    return short_description, long_description


def parse_docstring(
    docstring: str,
    *,
    key_sections: t.Iterable[str],
) -> data_models.Docstring:
    """
    Parses a docstring into

    - summary
    - details
    - sections

    Any sections listed in `key_sections` will be parsed as key-value pairs and
    returned as sections. Any remaining sections will be re-joined into the
    details.

    Any sections listed in `key_sections` that are not present in the docstring
    will be imputed as empty.
    """
    # Remove consistent indentation
    docstring = textwrap.dedent(docstring).strip()

    # Split the docstring into (summary + details) and sections
    summary_and_details, sections = split_docstring_into_sections(docstring)

    # Split into summary and details
    summary, details = parse_details(summary_and_details)
    del summary_and_details

    if details is None:
        details = ""

    # Find and parse all key-value sections
    key_sections = set(key_sections)
    key_value_sections: dict[str, dict[str, str]] = {}

    for section_name, section_contents in sections.items():
        # Key section?
        normalized_section_name = section_name.lstrip("#").strip().lower()

        if normalized_section_name in key_sections:
            key_value_sections[normalized_section_name] = parse_key_value_section(
                section_contents
            )

        # Text section
        else:
            details += f"\n\n{section_name}\n\n{section_contents}"

    # Don't keep around an empty details section
    details = details.strip()

    if not details:
        details = None

    # Make sure all requested sections are present
    missing_sections = key_sections - key_value_sections.keys()

    for missing_section in missing_sections:
        key_value_sections[missing_section] = {}

    # Done
    return data_models.Docstring(
        summary=summary,
        details=details,
        key_sections=key_value_sections,
    )


@t.overload
def parse(
    obj: data_models.ModuleObjectType,
    *,
    owner: data_models.ModuleDocs | None = None,
) -> data_models.ModuleDocs: ...


@t.overload
def parse(  # type: ignore (not mutually exclusive with the `Callable` overload)
    obj: data_models.ClassObjectType,
    *,
    owner: data_models.ModuleDocs | data_models.ClassDocs | None = None,
) -> data_models.ClassDocs: ...


@t.overload
def parse(
    obj: data_models.PropertyObjectType,
    *,
    owner: data_models.ClassDocs | None = None,
) -> data_models.PropertyDocs: ...


@t.overload
def parse(
    obj: data_models.FunctionObjectType,
    *,
    owner: data_models.ModuleDocs
    | data_models.ClassDocs
    | data_models.PropertyDocs
    | None = None,
) -> data_models.FunctionDocs: ...


def parse(
    obj: data_models.ModuleObjectType
    | data_models.FunctionObjectType
    | data_models.ClassObjectType
    | data_models.PropertyObjectType,
    *,
    owner: data_models.ModuleDocs
    | data_models.ClassDocs
    | data_models.PropertyDocs
    | None = None,
) -> (
    data_models.ModuleDocs
    | data_models.FunctionDocs
    | data_models.ClassDocs
    | data_models.PropertyDocs
):
    if isinstance(obj, types.ModuleType):
        return parse_module(obj, owner=t.cast(t.Any, owner))
    elif isinstance(obj, type):
        return parse_class(obj, owner=t.cast(t.Any, owner))
    elif isinstance(obj, property):
        return parse_property(obj, owner=t.cast(t.Any, owner))
    else:
        return parse_function(obj, owner=owner)


def parse_module(
    module: types.ModuleType,
    *,
    owner: data_models.ModuleDocs | None = None,
) -> data_models.ModuleDocs:
    # Parse the docstring
    docstring = inspect.getdoc(module)

    if docstring is None:
        parsed = data_models.Docstring(None, None, {})
        metadata = data_models.ModuleMetadata.from_dictionary({})
    else:
        parsed = parse_docstring(
            docstring,
            key_sections=["metadata"],
        )
        metadata = data_models.ModuleMetadata.from_dictionary(
            parsed.key_sections["metadata"]
        )

    # Collect all the public members
    module_dict = vars(module)
    try:
        public_member_names = module.__all__
    except AttributeError:

        def is_public(name: str, obj: object) -> bool:
            if name.startswith("_"):
                return False

            # Most modules are imported because they're required, not because
            # they're part of the public interface
            if isinstance(obj, types.ModuleType):
                # External modules are never public
                if obj.__package__ != module.__name__:
                    return False

                # Submodules may be public, but it's very rare and I can't think
                # of a reliable way to figure it out. Treating them all as
                # private has a much lower error rate than anything else I've
                # tried.
                return False

            # Classes and functions are considered public if they've been
            # defined in this module or a submodule
            if hasattr(obj, "__module__"):
                return (obj.__module__ + ".").startswith(module.__name__ + ".")

            # Some things, like TypeVars, are almost certainly not public
            if isinstance(obj, (t.TypeVar,)):
                return False

            return True

        public_member_names = [
            name for name, obj in module_dict.items() if is_public(name, obj)
        ]

    module_ = data_models.ModuleDocs(
        object=module,
        owner=owner,
        name=module.__name__.split(".")[-1],
        summary=parsed.summary,
        details=parsed.details,
        members={},
        metadata=metadata,
        deprecations=[],
    )

    for name in public_member_names:
        obj = module_dict[name]

        # Currently, we only support modules, classes and functions, so...
        if not isinstance(obj, (types.ModuleType, types.FunctionType, type)):
            continue

        member_docs = parse(obj, owner=module_)

        if member_docs.metadata.public:
            module_.add_member(member_docs)

    return module_


def parse_function(
    func: data_models.FunctionObjectType | classmethod | staticmethod,
    *,
    owner: data_models.ModuleDocs
    | data_models.ClassDocs
    | data_models.PropertyDocs
    | None = None,
) -> data_models.FunctionDocs:
    """
    Given a function, parse its signature and docstring into a `Function`
    object.
    """

    def parse_annotation(
        annotation: introspection.types.TypeAnnotation | type[inspect._empty],
    ) -> introspection.types.TypeAnnotation | data_models.Unset:
        if annotation is inspect.Parameter.empty:
            return data_models.UNSET

        # Recursive types can cause issues. In particular, we have `Jsonable`,
        # which is a recursive type, and `JsonDoc`, which references `Jsonable`.
        # So if a module imports `JsonDoc` but doesn't import `Jsonable`, it's
        # impossible to evaluate the forward reference `"Jsonable"` in that
        # module.
        #
        # A similar problem exists due to dataclasses: Every `Component`
        # subclass receives a constructor based on the type annotations in
        # `Component`, but of course the relevant imports aren't there.
        #
        # As a workaround, we'll make the missing names available in every
        # module.
        extra_globals = collections.ChainMap(
            {"Jsonable": Jsonable},
            vars(t),  # type: ignore
        )

        return introspection.typing.resolve_forward_refs(
            annotation,
            func.__module__,
            extra_globals=extra_globals,
            mode="ast",
            treat_name_errors_as_imports=True,
            strict=True,
        )

    is_class_method = is_static_method = False

    if isinstance(func, staticmethod):
        is_static_method = True
        func = t.cast(data_models.FunctionObjectType, func.__func__)
    elif isinstance(func, classmethod):
        is_class_method = True
        func = t.cast(data_models.FunctionObjectType, func.__func__)

    try:
        signature = inspect.signature(func)
    except ValueError:
        # Some builtins don't have an accessible signature. For now, we'll just
        # use a dummy.
        signature = inspect.Signature()

    # Parse the docstring
    docstring = inspect.getdoc(func)

    # Ignore the default docstring (inherited from `object`, I believe)
    if docstring == "Initialize self.  See help(type(self)) for accurate signature.":
        docstring = None

    if docstring is None:
        parsed = data_models.Docstring(
            summary=None, details=None, key_sections={"parameters": {}}
        )
        raises = []
        metadata = data_models.FunctionMetadata.from_dictionary({})
    else:
        parsed = parse_docstring(
            docstring,
            key_sections=[
                "parameters",
                "raises",
                "metadata",
            ],
        )

        # Add information about raised exceptions
        raises = list(parsed.key_sections["raises"].items())

        # Parse the metadata
        metadata = data_models.FunctionMetadata.from_dictionary(
            parsed.key_sections["metadata"]
        )

    # Check for deprecations
    unwrapped_func = inspect.unwrap(func)
    try:
        deprecations = deprecated_objects[unwrapped_func].copy()
    except KeyError:
        deprecations = []

    # Build the result
    function_ = data_models.FunctionDocs(
        object=func,
        owner=owner,
        name=func.__name__,
        parameters={},
        return_type=parse_annotation(signature.return_annotation),
        synchronous=not inspect.iscoroutinefunction(unwrapped_func),
        class_method=is_class_method,
        static_method=is_static_method,
        summary=parsed.summary,
        details=parsed.details,
        raises=raises,
        metadata=metadata,
        deprecations=deprecations,
    )

    # Parse the parameters
    for param_name, param in signature.parameters.items():
        # Skip private parameters. (It's rare to have private parameters, but
        # they do exist. Dataclasses in particular tend to have them.)
        if is_private_name(param_name):
            continue

        if param.default == inspect.Parameter.empty:
            param_default = data_models.UNSET
        else:
            param_default = param.default

        # Check for deprecations
        try:
            deprecations = deprecated_parameters[func][param_name].copy()
        except KeyError:
            deprecations = []

        param_ = data_models.ParameterDocs(
            owner=function_,
            name=param_name,
            type=parse_annotation(param.annotation),
            default=param_default,
            kw_only=param.kind == inspect.Parameter.KEYWORD_ONLY,
            collect_positional=param.kind == inspect.Parameter.VAR_POSITIONAL,
            collect_keyword=param.kind == inspect.Parameter.VAR_KEYWORD,
            description=None,
            deprecations=deprecations,
        )
        function_.parameters[param_name] = param_

    # Add any information learned about parameters from the docstring
    for param_name, param_details in parsed.key_sections["parameters"].items():
        try:
            result_param = function_.parameters[param_name]
        except KeyError:
            warnings.warn(
                f"The docstring for function `{func.__name__}` mentions a parameter `{param_name}` that does not exist in the function signature."
            )
            continue

        result_param.description = param_details

    return function_


def _parse_class_docstring_with_inheritance(
    cls: type,
    *,
    key_sections: t.Iterable[str],
    merge_key_section: t.Callable[
        [str, dict[str, str], dict[str, str]], dict[str, str]
    ] = lambda name, parent, child: {**parent, **child},
) -> data_models.Docstring:
    """
    Parses the docstring of a class in the same format as `parse_docstring`, but
    accounts for inheritance: Key-Value sections of all classes are merged, in a
    way that preserves child  over parent .
    """

    # Parse the docstring for this class
    raw_ = inspect.getdoc(cls)

    key_sections = set(key_sections)
    parsed_ = parse_docstring(
        "" if raw_ is None else raw_,
        key_sections=key_sections,
    )

    # Get the docstrings for the base classes
    base_: list[data_models.Docstring] = []

    for base in cls.__bases__:
        base_.append(
            _parse_class_docstring_with_inheritance(
                base,
                key_sections=key_sections,
                merge_key_section=merge_key_section,
            )
        )

    # Merge the docstrings
    result_sections: dict[str, dict[str, str]] = {}
    all_in_order = base_ + [parsed_]

    for cur_ in all_in_order:
        for section_name, cur_section in cur_.key_sections.items():
            result_section = result_sections.get(section_name, {})
            result_section = merge_key_section(
                section_name, result_section, cur_section
            )
            result_sections[section_name] = result_section

    parsed_.key_sections = result_sections

    # Done
    return parsed_


def parse_class(
    cls: type,
    *,
    owner: data_models.ClassOwnerType | None = None,
) -> data_models.ClassDocs:
    """
    Given a class, parse its signature an docstring into a `Class` object.
    """

    # Prepare a function for merging key sections
    def merge_key_section(
        name: str,
        parent: dict[str, str],
        child: dict[str, str],
    ) -> dict[str, str]:
        # Metadata is special: The parent doesn't matter, the child always wins.
        # This is to avoid unexpected situations, where a class would be e.g.
        # private, just because the parent is private.
        if name == "metadata":
            return child

        # Otherwise merge the two. Child values override parent values.
        return {**parent, **child}

    # Parse the docstring
    docstring = _parse_class_docstring_with_inheritance(
        cls,
        key_sections=[
            "attributes",
            "metadata",
        ],
        merge_key_section=merge_key_section,
    )

    # Check for deprecations
    try:
        deprecations = deprecated_objects[cls].copy()
    except KeyError:
        deprecations = []

    # Build the result. Functions have a backreference to the class , so
    # they will be added later.
    class_ = data_models.ClassDocs(
        object=cls,
        owner=owner,
        name=cls.__name__,
        members={},
        attributes={},
        summary=docstring.summary,
        details=docstring.details,
        metadata=data_models.ClassMetadata.from_dictionary(
            docstring.key_sections["metadata"]
        ),
        deprecations=deprecations,
    )

    # Parse the fields
    #
    # Make sure to add fields from base classes as well
    def add_fields(cls: type) -> None:
        # Processing `type` causes all sorts of problems, since it's the "root
        # metaclass". Best to just skip it.
        if cls is type:
            return

        # Chain to the base classes
        for base in cls.__bases__:
            add_fields(base)

        # Then process this class. This way locals fields override inherited
        # ones.
        for attribute_name, annotation in vars(cls).get("__annotations__", {}).items():
            # Skip `_: KW_ONLY` and `_private_attributes`
            if is_private_name(attribute_name):
                continue

            attribute_type = introspection.typing.resolve_forward_refs(
                annotation=annotation,
                context=cls.__module__,
                mode="ast",
                strict=True,
                treat_name_errors_as_imports=True,
            )

            field = data_models.AttributeDocs(
                owner=class_,
                name=attribute_name,
                type=attribute_type,
                default=data_models.UNSET,
                description=None,
                deprecations=[],
            )
            class_.attributes[attribute_name] = field

        # Properties are also fields
        for prop_name, prop in vars(cls).items():
            # Skip private attributes
            if is_private_name(prop_name):
                continue

            if isinstance(prop, property):
                prop_docs = data_models.PropertyDocs.from_property(prop, owner=class_)

                if prop_docs.metadata.public:
                    class_.members[prop_name] = prop_docs

    add_fields(cls)

    # Is this a dataclass? If so, get the fields from there
    if is_dataclass(cls):
        for dataclass_field in dataclasses.fields(cls):
            try:
                doc_field = class_.attributes[dataclass_field.name]
            except KeyError:
                # If it's not in the dict, it was deemed private. Skip it.
                continue

            # Default value?
            if dataclass_field.default is not dataclasses.MISSING:
                doc_field.default = dataclass_field.default

            # Default factory?
            elif dataclass_field.default_factory is not dataclasses.MISSING:
                doc_field.default = data_models.SENTINEL

    # Add any information learned about fields from the docstring
    for field_name, field_details in docstring.key_sections["attributes"].items():
        try:
            result_field = class_.attributes[field_name]
        except KeyError:
            # Warn about the missing field, unless there's a property with that
            # name
            if not isinstance(getattr(cls, field_name, None), property):
                warnings.warn(
                    f"The docstring for class `{cls.__name__}` mentions a field `{field_name}` that does not exist in the class."
                )
            continue

        result_field.description = field_details

    # Parse the functions
    #
    # Make sure to add functions from base classes as well
    seen_functions = set[str]()

    def add_functions(cls: type) -> None:
        # Skip well-known base classes like `Enum`
        if cls is enum.Enum:
            return

        # First process this class...
        for name, func in vars(cls).items():
            # Skip private methods
            if is_private_name(name):
                continue

            # Skip overridden methods
            if name in seen_functions:
                continue

            if (
                inspect.isfunction(func)
                or isinstance(func, classmethod)
                or isinstance(func, staticmethod)
            ):
                seen_functions.add(name)

                func_ = parse_function(func, owner=class_)

                if func_.metadata.public:
                    class_.members[name] = func_

        # ...then chain to the base classes
        for base in cls.__bases__:
            add_functions(base)

    add_functions(cls)

    # If `__init__` is missing documentation for any parameters, try to copy the
    # values from matching fields.
    init_method = class_.init_method
    if init_method is not None:
        for param in init_method.parameters.values():
            if param.description is not None:
                continue

            try:
                attr = class_.attributes[param.name]
            except KeyError:
                continue

            param.description = attr.description

    return class_


def parse_property(
    prop: property, owner: data_models.PropertyOwnerType | None
) -> data_models.PropertyDocs:
    assert prop.fget is not None

    if isinstance(owner, type):
        owner = data_models.ClassDocs.from_class(owner)

    getter = data_models.FunctionDocs.from_function(
        t.cast(data_models.FunctionObjectType, prop.fget)
    )
    summary = getter.summary
    details = getter.details
    deprecations = getter.deprecations
    metadata = getter.metadata

    if prop.fset is None:
        setter = None
    else:
        setter = data_models.FunctionDocs.from_function(
            t.cast(data_models.FunctionObjectType, prop.fset)
        )
        summary = summary or setter.summary
        details = details or setter.details
        deprecations = deprecations + setter.deprecations  # DON'T use `+=`!
        metadata = metadata or setter.metadata

    property_docs = data_models.PropertyDocs(
        object=prop,
        owner=owner,
        name=getter.name,
        getter=getter,
        setter=setter,
        summary=summary,
        details=details,
        metadata=data_models.PropertyMetadata.from_function_metadata(metadata),
        deprecations=deprecations,
    )

    getter.owner = property_docs
    if setter is not None:
        setter.owner = property_docs

    return property_docs


def is_private_name(name: str) -> bool:
    # Rules:
    # - `_foo` is private
    # - `__init__` is public
    # - `__foo__` is private
    if name == "__init__":
        return False

    return name.startswith("_")



import logging
import tarfile
import typing as t
from pathlib import Path


class AssetError(Exception):
    """
    Raised when an asset related issue occurs.
    """

    pass


class AssetManager:
    """
    Provides easy access to the project's assets. Assets are individually
    `.tar.xz` compressed and live in the `xz_dir` directory. This class allows
    accessing them via relative paths. Whenever an asset is needed it is
    extracted first, unless it already exists in the cache.

    The asset manager is versioned, to ensure that assets of different library
    versions don't clash. Future versions may use this to delete old, unused
    assets but this is not implemented yet.
    """

    def __init__(
        self,
        xz_dir: Path,
        cache_dir: Path,
        version: str,
    ) -> None:
        # The to the directory the .xz compressed assets are located in.
        #
        # Does not have to exist, though in that case obviously no assets can be
        # loaded.
        self.xz_dir = xz_dir

        # The path to the directory where extracted assets are stored. Does not
        # need to exist.
        self.cache_dir = cache_dir

        # Will be used to version extracted assets, so that future versions
        # don't have to extract them again.
        #
        # Must only consist of characters which are valid in a path.
        self._versioned_cache_dir = cache_dir / version

    def path_to_asset(
        self,
        path: Path,
        *,
        compressed: bool = True,
    ) -> Path:
        """
        Ensures that the given asset is extracted and returns a path to it. This
        is useful for assets that are not opened with `open_asset`. If you want
        to open the file use that function instead.

        Returns a path to where the extracted asset is located.

        ## Parameters

        path: The relative path to the asset.

        compressed: Whether the asset is compressed. If it is, the function will
            ensure that the asset is extracted before returning the path.
            Uncompressed assets are returned as-is.

        ## Raises

        ValueError: If the given path is not relative.

        FileNotFoundError: If there is no asset at the given path.

        AssetError: If the given asset cannot be extracted for any reason.
        """
        # Only allow relative paths, since they're supposed to be relative to
        # the asset directory.
        if path.is_absolute():
            raise ValueError("The asset's path must be relative")

        # Compressed assets need extracting
        if compressed:
            # If the file already exists, there is nothing else to do
            extracted_path = self._versioned_cache_dir / path

            if extracted_path.exists():
                return extracted_path

            # Make sure the target directory exists
            extracted_path.parent.mkdir(parents=True, exist_ok=True)

            # Extract the file
            compressed_path = self.xz_dir / path
            compressed_path = compressed_path.with_suffix(
                compressed_path.suffix + ".tar.xz"
            )

            logging.debug(f"Extracting asset `{path}` to `{extracted_path}`")

            try:
                with tarfile.open(compressed_path, "r:xz") as tar:
                    tar.extractall(extracted_path.parent)

            except FileNotFoundError:
                raise FileNotFoundError(f"There is no asset at `{path}`")

            except tarfile.TarError as e:
                raise AssetError(
                    f"Could not extract the asset from `{path}`: {e}"
                ) from e

            except OSError as e:
                raise AssetError(
                    f"Could not read the asset file at `{path}`: {e}"
                ) from e

        # Uncompressed can be used as-is
        elif not (self.xz_dir / path).exists():
            raise FileNotFoundError(f"There is no asset at `{path}`")

        # Done
        return extracted_path

    @t.overload
    def open_asset(
        self,
        path: Path,
        mode: t.Literal["r"],
        *,
        compressed: bool = True,
    ) -> t.IO[str]: ...

    @t.overload
    def open_asset(
        self,
        path: Path,
        mode: t.Literal["rb"],
        *,
        compressed: bool = True,
    ) -> t.IO[bytes]: ...

    def open_asset(
        self,
        path: Path,
        mode: t.Literal["r", "rb"] = "r",
        *,
        compressed: bool = True,
    ) -> t.IO:
        """
        Opens an asset file for reading.

        Returns a file-like object.

        ## Parameters

        path: The relative path to the asset.

        mode: The mode to open the file in. The default is `"r"`.

        compressed: Whether the asset is compressed. If it is, the function will
            ensure that the asset is extracted before opening it. Uncompressed
            assets are opened as-is.

        ## Raises

        ValueError: If the given path is not relative.

        FileNotFoundError: If there is no asset at the given path.

        AssetError: If the asset cannot be opened for any reason.
        """
        # Make sure the file exists.
        extracted_path = self.path_to_asset(
            path=path,
            compressed=compressed,
        )

        # The file exists now - open it.
        return extracted_path.open(mode)

    def get_cache_path(self, path: Path) -> Path:
        """
        Returns where the manager would cache the extracted path, if it was to
        be extracted. This does not actually extract anything, nor does it check
        whether the archive for this asset exists.

        You can use this function if you want to dump your own cached files in
        the asset manager's cache directory.

        After this function returns, the parent directory of the returned path
        will exist.

        ## Raises

        ValueError: If the given path is not relative.

        AssetError: If the containing directory of the returned path could not
            be created.
        """
        # Only allow relative paths, since they're supposed to be relative to
        # the asset directory.
        if path.is_absolute():
            raise ValueError("The asset's path must be relative")

        # Build the result path
        result = self._versioned_cache_dir / path

        # Make sure the parent directory exists
        try:
            result.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            raise AssetError(
                f"Could not create the containing directory of `{result}`: {e}"
            ) from e

        # Done
        return result



from __future__ import annotations

import re
import typing as t

import introspection

__all__ = [
    "mark_as_private",
    "mark_constructor_as_private",
    "insert_links_into_markdown",
]


Class = t.TypeVar("Class", bound=type)
ClassOrFunction = t.TypeVar("ClassOrFunction", bound=t.Callable)


def mark_as_private(obj: ClassOrFunction) -> ClassOrFunction:
    """
    Adds `public: False` to the `## Metadata` section of the given object's
    docstring.
    """

    if obj.__doc__ is None:
        obj.__doc__ = "## Metadata"
    elif "## Metadata" not in obj.__doc__:
        obj.__doc__ += "\n## Metadata"

    obj.__doc__ += "\n`public`: False"

    return obj


def mark_constructor_as_private(cls: Class) -> Class:
    """
    Calls `mark_as_private` on the class's `__init__` method. (If the `__init__`
    method is inherited from a parent class, a new one will be created.)
    """

    try:
        constructor = vars(cls)["__init__"]
    except KeyError:

        @introspection.mark.does_not_alter_signature
        def constructor(self, *args, **kwargs):
            super(cls, self).__init__(*args, **kwargs)  # type: ignore (wtf?)

        introspection.add_method_to_class(constructor, cls, "__init__")

    mark_as_private(constructor)  # type: ignore (wtf?)

    return cls


_CODE_BLOCK_REGEX = re.compile(r"(?!$)(.*?)(```.*?```|$)", flags=re.S)
_NAME_TO_LINK_REGEX = re.compile(
    r"`([a-zA-Z_.]+)`(?!\])"
)  # TODO: Why does this specifically look for a "]" character?


def insert_links_into_markdown(
    markdown: str,
    url_for_name: t.Callable[[str], str | None],
) -> str:
    r"""
    Looks for references to functions/classes/etc and turns them into
    hyperlinks.

    References are identifiers enclosed in backticks, for example:

    - `` `MyClass` ``
    - `` `my_module.MyClass` ``

    Every found reference is passed to the `url_for_name` function. If the
    function returns `None`, no hyperlink will be created.
    """

    def repl(match: re.Match) -> str:
        url = url_for_name(match.group(1))

        if url is None:
            return match.group()

        return f"[{match.group()}]({url})"

    # We want to look for single-line text like `MyClass` and turn it
    # into a link. The problem is that such text might be appear inside
    # of a code block (in a comment). So we'll search for code blocks,
    # and only apply the substituation in the text before the code
    # block.
    chunks = list[str]()

    for match_ in _CODE_BLOCK_REGEX.finditer(markdown):
        text, code_block = match_.groups()

        chunks.append(_NAME_TO_LINK_REGEX.sub(repl, text))
        chunks.append(code_block)

    return "".join(chunks)



from __future__ import annotations

import abc
import dataclasses
import types
import typing as t

import introspection.types
import introspection.typing
import typing_extensions as te

from ..deprecations import Deprecation
from . import parsers

__all__ = [
    "Unset",
    "UNSET",
    "Sentinel",
    "SENTINEL",
    "CommonMetadata",
    "Deprecation",
    "Docstring",
    "ModuleDocs",
    "ParameterDocs",
    "FunctionDocs",
    "AttributeDocs",
    "ClassDocs",
    "PropertyDocs",
]


AnyDocs = t.Union[
    "ModuleDocs",
    "ClassDocs",
    "FunctionDocs",
    "PropertyDocs",
    "AttributeDocs",
    "ParameterDocs",
]


class Unset:
    pass


UNSET = Unset()


class Sentinel:
    pass


SENTINEL = Sentinel()


@dataclasses.dataclass
class CommonMetadata:
    """
    Some metadata such as whether an object is public or not is shared between
    different types of objects. This class is used to hold that metadata.
    """

    # Whether the object is meant to be used by users of the library, or if it's
    # an internal implementation detail.
    public: bool = True

    # If `True`, this object is not yet ready for public use. Its API may change
    # between even patch releases.
    experimental: bool = False

    # The version when this object was first added.
    added_in_version: str | None = None

    # Contains all `key: value` pairs that don't correspond to known fields
    extras: dict[str, str] = dataclasses.field(default_factory=dict)

    @staticmethod
    def _parse_bool(value: str) -> bool:
        """
        Attempts to parse a boolean value from metadata.

        ## Raises

        `ValueError`: If the key is invalid.
        """
        # Postprocess the value
        if isinstance(value, str):
            value = value.strip()

        # Recognized strings
        if value == "True":
            return True

        if value == "False":
            return False

        # Invalid value
        raise ValueError(f"Cannot parse {value!r} as a boolean")

    @staticmethod
    def _parse_literal(
        metadata: dict[str, str],
        key_name: str,
        options: t.Set[str],
        default_value: str | None,
    ) -> str:
        """
        Attempts to parse a literal value from metadata.

        ## Raises

        `ValueError`: If the key is missing or invalid.
        """

        # Try to get the value
        try:
            raw = metadata[key_name]
        except KeyError:
            # No value provided
            if default_value is None:
                raise ValueError(f"Missing value for `{key_name}` in metadata")

            return default_value

        # Postprocess the value
        if isinstance(raw, str):
            raw = raw.strip()

        # Check if the value is valid
        if raw not in options:
            raise ValueError(f'Invalid value for `{key_name}` in metadata: "{raw}"')

        return raw

    @classmethod
    def from_dictionary(cls, metadata: dict[str, t.Any]) -> te.Self:
        """
        Parses a `CommonMetadata` object from a dictionary. This is useful for
        parsing metadata from a docstring key section.
        """

        kwargs = {}
        extras = {}

        type_hints = t.get_type_hints(cls)

        for key, value in metadata.items():
            try:
                annotation = type_hints[key]
            except KeyError:
                # Unknown field
                extras[key] = value
                continue

            try:
                if annotation is bool:
                    parsed_value = cls._parse_bool(value)
                elif annotation is str:
                    parsed_value = value
                else:
                    raise NotImplementedError(
                        f"Can't parse values of type {annotation} yet"
                    )
            except ValueError:
                raise ValueError(f"Invalid value for {key!r}: {value!r}")

            kwargs[key] = parsed_value

        # Construct the result
        return cls(**kwargs, extras=extras)


@dataclasses.dataclass
class FunctionMetadata(CommonMetadata):
    decorator: bool = False


@dataclasses.dataclass
class ClassMetadata(CommonMetadata):
    pass


@dataclasses.dataclass
class ModuleMetadata(CommonMetadata):
    pass


@dataclasses.dataclass
class PropertyMetadata(CommonMetadata):
    @staticmethod
    def from_function_metadata(
        function_metadata: FunctionMetadata,
    ) -> PropertyMetadata:
        return PropertyMetadata(
            public=function_metadata.public,
            experimental=function_metadata.experimental,
            added_in_version=function_metadata.added_in_version,
            extras=function_metadata.extras.copy(),
        )


@dataclasses.dataclass
class Docstring:
    """
    A generic docstring object.

    Docstrings are split into multiple sections: The **summary** is a brief,
    one-line description of the object. This is intended to be displayed right
    next to the object's name in a list of objects for example.

    The **details** section is a more in-depth explanation of the object. This
    may span multiple paragraphs and gives an explanation of the object

    Finally, **key_sections** are sections which consist entirely of `key:
    value` pairs. These can be used for raised exceptions, parameters, and
    similar.
    """

    summary: str | None
    details: str | None

    key_sections: dict[str, dict[str, str]]

    @staticmethod
    def from_string(
        docstring: str,
        *,
        key_sections: t.Iterable[str],
    ) -> Docstring:
        return parsers.parse_docstring(
            docstring,
            key_sections=key_sections,
        )


DirectChildrenType = t.TypeVar("DirectChildrenType", bound="_Docs")
IndirectChildrenType = t.TypeVar("IndirectChildrenType", bound="_Docs")
OwnerType = t.TypeVar("OwnerType", bound="_Docs")
ObjectType = t.TypeVar("ObjectType")


@dataclasses.dataclass
class _Docs(abc.ABC, t.Generic[OwnerType, DirectChildrenType, IndirectChildrenType]):
    """
    `_Docs[OwnerType, DirectChildrenType, IndirectChildrenType]`

    Base class for all Docs.
    """

    # These two attributes contain the *public* name/location of this object,
    # *not* the location where it was defined. For example, the `owner` of
    # `requests.Session` would be the `requests` module, even though that class
    # is defined in the `requests.session` sub-module.
    owner: OwnerType | None = dataclasses.field(repr=False)
    name: str

    deprecations: list[Deprecation]

    @t.overload
    def iter_children(
        self,
        *,
        recursive: t.Literal[False],
        include_self: t.Literal[False] = False,
    ) -> t.Iterable[DirectChildrenType]: ...

    @t.overload
    def iter_children(
        self,
        *,
        recursive: t.Literal[True],
        include_self: t.Literal[False] = False,
    ) -> t.Iterable[DirectChildrenType | IndirectChildrenType]: ...

    @t.overload
    def iter_children(
        self,
        *,
        recursive: t.Literal[False],
        include_self: t.Literal[True],
    ) -> t.Iterable[te.Self | DirectChildrenType]: ...

    @t.overload
    def iter_children(
        self,
        *,
        recursive: t.Literal[True],
        include_self: t.Literal[True],
    ) -> t.Iterable[te.Self | DirectChildrenType | IndirectChildrenType]: ...

    def iter_children(
        self, *, recursive: bool, include_self: bool = False
    ) -> t.Iterable[te.Self | DirectChildrenType | IndirectChildrenType]:
        """
        Yields all child docs of this docs object. (For example, a
        `FunctionDocs` will yield its `ParameterDocs`.)
        """

        if include_self:
            yield self

        yield from self._iter_children(recursive=recursive)

    @abc.abstractmethod
    def _iter_children(
        self, *, recursive: bool
    ) -> t.Iterable[DirectChildrenType | IndirectChildrenType]:
        raise NotImplementedError

    @abc.abstractmethod
    def transform_docstrings(self, transform: t.Callable[[str], str]) -> None:
        """
        Applies a transformation function to all "docstrings" of this docs
        object. (This is useful because some docs have a `description`, while
        others have a `summary` and `details`.)
        """
        raise NotImplementedError


@dataclasses.dataclass
class _SummaryAndDetailsMixin:
    summary: str | None
    details: str | None

    def transform_docstrings(self, transform: t.Callable[[str], str]) -> None:
        if self.summary is not None:
            self.summary = transform(self.summary)

        if self.details is not None:
            self.details = transform(self.details)


@dataclasses.dataclass
class _DescriptionMixin:
    description: str | None

    def transform_docstrings(self, transform: t.Callable[[str], str]) -> None:
        if self.description is not None:
            self.description = transform(self.description)


class _DocsWithoutChildren(_Docs[OwnerType, te.Never, te.Never]):
    def _iter_children(self, *, recursive: bool) -> t.Iterable[te.Never]:
        return ()


@dataclasses.dataclass
class _ObjectDocs(
    _SummaryAndDetailsMixin,
    _Docs[OwnerType, DirectChildrenType, IndirectChildrenType],
    t.Generic[OwnerType, DirectChildrenType, IndirectChildrenType, ObjectType],
):
    """
    `_ObjectDocs[OwnerType, DirectChildrenType, IndirectChildrenType, ObjectType]`

    Base class for everything that's an object, i.e. exists at runtime.
    (i.e. modules, classes, functions, properties, ...)
    """

    object: ObjectType

    @property
    def full_name(self) -> str:
        """
        The "full name" of this object, in other words, how users are expected
        to access it. (For example, "requests.Session")
        """
        parts = list[str]()

        obj = self
        while obj is not None:
            parts.append(obj.name)
            obj = obj.owner

        parts.reverse()
        return ".".join(parts)


@dataclasses.dataclass
class _ScopeDocs(
    _ObjectDocs[OwnerType, DirectChildrenType, IndirectChildrenType, ObjectType]
):
    """
    `_ScopeDocs[OwnerType, DirectChildrenType, IndirectChildrenType, ObjectType]`

    Base class for objects that store other objects as attributes, like modules
    and classes.
    """

    members: dict[str, DirectChildrenType] = dataclasses.field(repr=False)

    def add_member(
        self,
        member: DirectChildrenType,
        *,
        name: str | None = None,
    ) -> None:
        """
        Adds the given object as a member. You can use this method to add
        objects that were incorrectly assumed to be private.
        """
        if name is None:
            name = member.name

        self.members[name] = member

    def _iter_children(
        self, *, recursive: bool
    ) -> t.Iterable[DirectChildrenType | IndirectChildrenType]:
        for member in self.members.values():
            yield member

            if recursive:
                yield from member._iter_children(recursive=True)  # type: ignore


ModuleOwnerType: te.TypeAlias = "ModuleDocs"
ModuleDirectChildrenType = t.Union["ModuleDocs", "ClassDocs", "FunctionDocs"]
ModuleIndirectChildrenType = t.Union["PropertyDocs", "AttributeDocs", "ParameterDocs"]
ModuleObjectType = types.ModuleType


@dataclasses.dataclass
class ModuleDocs(
    _ScopeDocs[
        ModuleOwnerType,
        ModuleDirectChildrenType,
        ModuleIndirectChildrenType,
        ModuleObjectType,
    ]
):
    metadata: ModuleMetadata

    @staticmethod
    def from_module(
        module: types.ModuleType,
        *,
        owner: ModuleDocs | None = None,
    ) -> ModuleDocs:
        """
        Parses a `ModuleDocs` object from a module object.
        """
        return parsers.parse_module(module, owner=owner)

    def _iter_children(
        self, *, recursive: bool
    ) -> t.Iterable[ModuleDirectChildrenType | ModuleIndirectChildrenType]:
        for member in self.members.values():
            yield member

            if recursive:
                yield from member._iter_children(recursive=True)

    def add_member(
        self,
        member: ModuleObjectType
        | ClassObjectType
        | FunctionObjectType
        | ModuleDirectChildrenType,
        *,
        name: str | None = None,
    ) -> None:
        if not isinstance(member, _Docs):
            member = parsers.parse(member, owner=self)

        return super().add_member(member, name=name)


FunctionOwnerType = t.Union["ModuleDocs", "ClassDocs", "PropertyDocs"]
FunctionDirectChildrenType: te.TypeAlias = "ParameterDocs"
FunctionIndirectChildrenType = te.Never
FunctionObjectType: te.TypeAlias = t.Callable


@dataclasses.dataclass
class FunctionDocs(
    _ObjectDocs[
        FunctionOwnerType,
        FunctionDirectChildrenType,
        FunctionIndirectChildrenType,
        FunctionObjectType,
    ]
):
    parameters: dict[str, ParameterDocs]
    return_type: introspection.types.TypeAnnotation | Unset
    synchronous: bool

    class_method: bool
    static_method: bool

    raises: list[tuple[str, str]]  # type, description

    metadata: FunctionMetadata

    @property
    def has_implicit_first_parameter(self) -> bool:
        """
        Returns `True` for instance methods, class methods, property getters and
        property setters.

        Note that this may incorrectly return `False` if `owner` is set to
        `None`.
        """
        if self.static_method:
            return False

        if self.class_method:
            return True

        return isinstance(self.owner, (ClassDocs, PropertyDocs))

    @staticmethod
    def from_function(
        func: FunctionObjectType,
        *,
        owner: FunctionOwnerType | None = None,
    ) -> FunctionDocs:
        """
        Parses a `FunctionDocs` object from a function or method. This takes
        both the function's docstring as well as its signature and type hints
        into account.
        """
        return parsers.parse_function(func, owner=owner)

    def _iter_children(
        self, *, recursive: bool
    ) -> t.Iterable[FunctionDirectChildrenType | FunctionIndirectChildrenType]:
        yield from self.parameters.values()


@dataclasses.dataclass
class ParameterDocs(_DescriptionMixin, _DocsWithoutChildren["FunctionDocs"]):
    type: introspection.types.TypeAnnotation | Unset
    default: object | Unset

    kw_only: bool

    collect_positional: bool
    collect_keyword: bool


@dataclasses.dataclass
class AttributeDocs(_DescriptionMixin, _DocsWithoutChildren["ClassDocs"]):
    type: introspection.types.TypeAnnotation | Unset
    default: object | Unset


PropertyOwnerType: te.TypeAlias = "ClassDocs"
PropertyDirectChildrenType: te.TypeAlias = "FunctionDocs"
PropertyIndirectChildrenType = t.Union[
    FunctionDirectChildrenType, FunctionIndirectChildrenType
]
PropertyObjectType = property


@dataclasses.dataclass
class PropertyDocs(
    _ObjectDocs[
        PropertyOwnerType,
        PropertyDirectChildrenType,
        PropertyIndirectChildrenType,
        PropertyObjectType,
    ]
):
    getter: FunctionDocs
    setter: FunctionDocs | None

    metadata: PropertyMetadata

    @staticmethod
    def from_property(
        prop: PropertyObjectType,
        *,
        owner: PropertyOwnerType | None = None,
    ) -> PropertyDocs:
        return parsers.parse_property(prop, owner=owner)

    def _iter_children(
        self, *, recursive: bool
    ) -> t.Iterable[PropertyDirectChildrenType | PropertyIndirectChildrenType]:
        yield self.getter
        if recursive:
            yield from self.getter._iter_children(recursive=True)

        if self.setter is not None:
            yield self.setter
            if recursive:
                yield from self.setter._iter_children(recursive=True)


ClassOwnerType: te.TypeAlias = "ModuleDocs"
ClassDirectChildrenType = t.Union["AttributeDocs", "PropertyDocs", "FunctionDocs"]
ClassIndirectChildrenType = t.Union[
    PropertyDirectChildrenType,
    PropertyIndirectChildrenType,
    FunctionDirectChildrenType,
    FunctionIndirectChildrenType,
]
ClassObjectType = type


@dataclasses.dataclass
class ClassDocs(
    _ScopeDocs[
        ClassOwnerType,
        ClassDirectChildrenType,
        ClassIndirectChildrenType,
        ClassObjectType,
    ]
):
    attributes: dict[str, AttributeDocs]

    metadata: ClassMetadata

    @property
    def properties(self) -> t.Mapping[str, PropertyDocs]:
        return {
            name: docs
            for name, docs in self.members.items()
            if isinstance(docs, PropertyDocs)
        }

    @property
    def functions(self) -> t.Mapping[str, FunctionDocs]:
        return {
            name: docs
            for name, docs in self.members.items()
            if isinstance(docs, FunctionDocs)
        }

    @property
    def init_method(self) -> FunctionDocs | None:
        """
        Returns the `FunctionDocs` for this class's `__init__` method, if it has
        one.
        """
        try:
            init_method = self.members["__init__"]
        except KeyError:
            pass
        else:
            if isinstance(init_method, FunctionDocs):
                return init_method

        return None

    @staticmethod
    def from_class(
        typ: ClassObjectType,
        *,
        owner: ClassOwnerType | None = None,
    ) -> ClassDocs:
        """
        Parses a `ClassDocs` object from a class. This takes both the class's
        docstring as well as its methods and attributes into account.
        """
        return parsers.parse_class(typ, owner=owner)

    def _iter_children(
        self, *, recursive: bool
    ) -> t.Iterable[ClassDirectChildrenType | ClassIndirectChildrenType]:
        yield from super()._iter_children(recursive=recursive)
        yield from self.attributes.values()

    def add_member(
        self,
        member: FunctionObjectType | PropertyObjectType | ClassDirectChildrenType,
        *,
        name: str | None = None,
    ) -> None:
        if not isinstance(member, _Docs):
            member = parsers.parse(member, owner=self)

        return super().add_member(member, name=name)



import dataclasses
import warnings

import imy.deprecations


def test_deprecated_class():
    @imy.deprecations.deprecated(since="1.1", replacement="Bar")
    class Foo:
        pass

    with warnings.catch_warnings(record=True) as warnings_list:
        Foo()

    assert len(warnings_list) == 1


def test_deprecated_function_and_parameter():
    @imy.deprecations.deprecated(
        since="1.2",
        will_be_removed_in_version="1.5",
        description="It causes undefined behavior",
    )
    @imy.deprecations.parameter_renamed(
        since="1.3", old_name="old_param", new_name="param"
    )
    def foo(param: int):
        return param

    with warnings.catch_warnings(record=True) as warnings_list:
        assert foo(3) == 3

    assert len(warnings_list) == 1

    with warnings.catch_warnings(record=True) as warnings_list:
        assert foo(old_param=5) == 5  # type: ignore

    assert len(warnings_list) == 2


def test_parameter_remapping():
    @imy.deprecations.parameter_remapped(
        since="1.3",
        old_name="old_param",
        new_name="param",
        remap=str,
    )
    def foo(param: str):
        return param

    with warnings.catch_warnings(record=True) as warnings_list:
        assert foo("hi") == "hi"

    assert len(warnings_list) == 0

    with warnings.catch_warnings(record=True) as warnings_list:
        assert foo(old_param=5) == "5"  # type: ignore

    assert len(warnings_list) == 1


def test_dataclass_parameter_deprecation():
    @imy.deprecations.parameter_renamed(since="0.9.3", old_name="foo", new_name="bar")
    @dataclasses.dataclass
    class MyClass:
        bar: int

    assert MyClass(foo=5).bar == 5  # type: ignore


def test_stacktrace():
    @imy.deprecations.parameter_renamed(since="0.9.3", old_name="old", new_name="new")
    @imy.deprecations.parameter_renamed(since="0.9.3", old_name="foo", new_name="bar")
    def my_func(*, bar, new):
        pass

    with warnings.catch_warnings(record=True) as warnings_list:
        exec("""my_func(old=3, foo='hi')""", {"my_func": my_func})

    assert len(warnings_list) == 2
    for warning in warnings_list:
        assert warning.filename == "<string>"
        assert warning.lineno == 1



import asyncio
import threading
import typing as t

# Try to import the built-in TaskGroup from asyncio, falling back to the bundled
# backport if it's not available.
try:
    from asyncio import (
        BaseExceptionGroup as BaseExceptionGroup,  # type: ignore (Handled below)
    )
    from asyncio import (
        ExceptionGroup as ExceptionGroup,  # type: ignore (Handled below)
    )
    from asyncio import (
        TaskGroup as TaskGroup,  # type: ignore (Handled below)
    )
except ImportError:
    from ._task_group import (
        BaseExceptionGroup as BaseExceptionGroup,
    )
    from ._task_group import (
        ExceptionGroup as ExceptionGroup,
    )
    from ._task_group import (
        TaskGroup as TaskGroup,
    )

T = t.TypeVar("T")

T = t.TypeVar("T")
IN = t.TypeVar("IN")
OUT = t.TypeVar("OUT")


async def collect(
    values: t.AsyncIterable[IN],
    *,
    limit: int | None = None,
) -> list[IN]:
    """
    Collects the values from an asynchronous iterable into a list.

    If `limit` is set, the result will contain at most the first `limit` values.

    ## Parameters

    `values`: The asynchronous iterable to collect values from.

    `limit`: The maximum number of values to collect. If `None`, all values will
        be collected.
    """
    result: list[IN] = []

    async for value in values:
        result.append(value)

        if len(result) == limit:
            break

    return result


async def amap(
    func: t.Callable[[IN], t.Awaitable[OUT]],
    values: t.Iterable[IN],
    *,
    concurrency: int = 10,
) -> t.AsyncIterable[OUT]:
    """
    Asynchronously maps `func` over `values`, with at most `concurrency` items
    being processed simultaneously.

    ## Parameters

    `func`: The function to apply to each value.

    `values`: The values to apply the function to.

    `concurrency`: The maximum number of items to process simultaneously.
    """
    # TODO: Support async iterables for the input

    # Processes a single value. Returns both the index and the result.
    async def process_single(index: int, value: IN) -> tuple[int, OUT]:
        result = await func(value)
        return index, result

    # Yield values as they become available - however, make sure to raise asap
    # if any task fails
    to_do: list[tuple[int, IN]] = list(enumerate(values))
    in_flight: set[asyncio.Task[tuple[int, OUT]]] = set()

    n_already_yielded: int = 0
    not_yet_yielded: dict[int, OUT] = {}

    while True:
        # Spawn tasks until the limit is reached
        while len(in_flight) < concurrency:
            try:
                index, value = to_do.pop(0)
            except IndexError:
                break

            in_flight.add(
                asyncio.create_task(
                    process_single(
                        index,
                        value,
                    )
                )
            )

        # Done?
        if not in_flight and not to_do:
            break

        # Wait for any task to complete
        done, in_flight = await asyncio.wait(
            in_flight,
            return_when=asyncio.FIRST_COMPLETED,
        )

        # Add any completed tasks to the results
        for task in done:
            # Pass-through exceptions
            try:
                index, result = task.result()
            except Exception:
                for task in in_flight:
                    task.cancel()

                raise

            # Store the result
            assert index >= n_already_yielded, (
                index,
                n_already_yielded,
                not_yet_yielded,
            )
            not_yet_yielded[index] = result

        # Yield any results that are now available
        while True:
            try:
                value = not_yet_yielded.pop(n_already_yielded)
            except KeyError:
                break
            else:
                n_already_yielded += 1
                yield value


async def iterator_to_thread(
    sync_iterable: t.Iterable[T],
    *,
    batch_size: int,
) -> t.AsyncIterable[T]:
    """
    Given a potentially slow, synchronous iterator, returns an equivalent
    asynchronous iterator.

    The synchronous iterator will be run in a separate thread, and the results
    retrieved asynchronously. To decrease overhead, values from the synchronous
    iterator are fetched in batches. This means that values won't be yielded
    until `batch_size` values have been collected, or the iterator is exhausted.
    If this behavior isn't desirable in your application, set `batch_size` to
    `1`.

    Exceptions from the synchronous iterator will be propagated.
    """
    assert batch_size > 0, "Batch size must be positive"

    mainloop = asyncio.get_event_loop()
    results: asyncio.Queue[
        tuple[t.Literal["values"], list[T]]
        | tuple[t.Literal["error"], Exception]
        | tuple[t.Literal["done"]]
    ] = asyncio.Queue()
    stop_requested = False

    # The worker function continuously fills the queue with results. It is
    # limited by the buffer size.
    def worker():
        sync_iterator = iter(sync_iterable)
        batch: list[T] = []

        while True:
            # Try to get some values from the iterator
            try:
                batch.append(next(sync_iterator))

            # Done?
            except StopIteration:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                mainloop.call_soon_threadsafe(results.put_nowait, ("done",))
                return

            # Handle errors
            except Exception as e:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                mainloop.call_soon_threadsafe(results.put_nowait, ("error", e))
                return

            # If the batch is full, pass it on
            if len(batch) >= batch_size:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                batch = []

            # If the stop was requested, pass on the remaining values
            if stop_requested:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                mainloop.call_soon_threadsafe(results.put_nowait, ("done",))
                return

    worker_thread = threading.Thread(target=worker, daemon=True)

    # The remainder of the function has to be wrapped to ensure the thread is
    # stopped if anything goes wrong.
    try:
        worker_thread.start()

        while True:
            # Wait for the next result
            result = await results.get()

            # Yield the values
            if result[0] == "values":
                for value in result[1]:
                    yield value

            # Handle errors
            elif result[0] == "error":
                raise result[1]

            # Done
            elif result[0] == "done":
                return

            # Invalid
            else:
                assert False, f"Received invalid result `{result!r}`"

    finally:
        stop_requested = True


async def multitask(*coroutines: t.Awaitable[t.Any]) -> None:
    """
    **This function is deprecated. Use task groups instead.**

    Run multiple coroutines concurrently. When either completes, all others are
    canceled as well. If any coroutine raises an exception, they also all stop,
    and the exception is re-raised.

    This effectively allows you to run multiple streams of code at once,
    without having to worry about some of them living on when they're supposed
    to finish.

    If no coroutines are provided, the function returns immediately.

    Warning: While most exceptions are propagated, `asyncio.CancelledError` is
        not. That's because this exception is raised under normal circumstances
        simply when this function shuts down a coroutine.

    ## Raises

    `BaseException`: Any exceptions raised by the coroutines.
    """
    # Special case: If no coroutines are provided, the function is finished.
    # Avoid all of the overhead below, especially since some of the functions
    # probably aren't too happy about empty sets.
    if not coroutines:
        return

    # Create a task for each coroutine
    tasks = {asyncio.create_task(coro) for coro in coroutines}  # type: ignore

    # Run them all
    try:
        await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )

    # Make sure all tasks are canceled prior to returning
    finally:
        # Cancel any pending tasks
        for task in tasks:
            if not task.done():
                task.cancel()

        # Wait for all tasks to be canceled
        exception_to_raise: BaseException | None = None

        for task in tasks:
            try:
                await task
            except asyncio.CancelledError:
                # CancelledErrors are to be expected. They don't signal that
                # something has gone wrong, but simply that the task was
                # canceled, likely by this very function.
                pass
            except BaseException as err:
                exception_to_raise = err

        # Propagate any exceptions
        if exception_to_raise is not None:
            raise exception_to_raise


async def to_daemon_thread(
    func: t.Callable[..., T],
    *args: t.Any,
    **kwargs: t.Any,
) -> T:
    """
    Like `asyncio.to_thread`, but the thread is a daemon. This means that the
    thread will be automatically stopped when no other non-daemon threads are
    running.
    """
    loop = asyncio.get_running_loop()
    future = asyncio.Future()

    # Create a function to run in the thread. It will release the future once
    # done.
    def run() -> None:
        try:
            result = func(*args, **kwargs)
            loop.call_soon_threadsafe(future.set_result, result)
        except Exception as e:
            loop.call_soon_threadsafe(future.set_exception, e)

    # Get busy
    thread = threading.Thread(target=run, daemon=True)
    thread.start()

    # Wait for the result
    return await future



import collections
import dataclasses
import functools
import inspect
import types
import typing as t
import warnings

import introspection

__all__ = [
    "configure",
    "deprecated",
    "parameter_renamed",
    "parameter_remapped",
    "warn",
    "warn_parameter_renamed",
]


C = t.TypeVar("C", bound=t.Callable)


@dataclasses.dataclass
class Deprecation:
    since_version: str
    message: str
    will_be_removed_in_version: str | None = None


# Global data structures containing information that's useful for
# `imy.docstrings`
deprecated_objects = collections.defaultdict[
    t.Callable | types.ModuleType | property, list[Deprecation]
](list)
deprecated_parameters = collections.defaultdict[
    t.Callable, collections.defaultdict[str, list[Deprecation]]
](lambda: collections.defaultdict(list))


# Configs. Each config is associated with a module. This allows different
# modules to customize how their deprecation warnings are displayed.
class Config:
    def __init__(
        self,
        project_name: str,
        modules_skipped_in_stacktrace: t.Container[str] | None = None,
        warning_class: type[Warning] = DeprecationWarning,
        name_for_object: t.Callable[[t.Callable], str] = lambda obj: obj.__name__,
    ):
        if modules_skipped_in_stacktrace is None:
            modules_skipped_in_stacktrace = {project_name}

        self.project_name = project_name
        self.modules_skipped_in_stacktrace = modules_skipped_in_stacktrace
        self.warning_class = warning_class
        self.name_for_object = name_for_object

    def make_parameter_deprecation_message(
        self,
        old_name: str,
        new_name: str,
        callable_: t.Callable | str | None = None,
    ) -> str:
        if callable_ is None:
            of_what = ""
        else:
            if not isinstance(callable_, str):
                callable_ = self.name_for_object(callable_)

            of_what = f" of `{callable_}`"

        return f"The `{old_name}` parameter{of_what} has been renamed. Use `{new_name}` instead."

    def add_versions(
        self,
        message: str,
        *,
        since: str | None = None,
        will_be_removed_in_version: str | None = None,
    ) -> str:
        if since is None:
            intro = ""
        else:
            intro = f"Deprecated since {self.project_name} version {since}"

        if will_be_removed_in_version is not None:
            if intro:
                intro += f"; to be removed in version {will_be_removed_in_version}"
            else:
                intro = f"To be removed in version {will_be_removed_in_version}"

        return f"{intro}: {message}"

    def warn(self, message: str) -> None:
        # Find the first stack frame outside of the module that's issuing the
        # warning. (Passing the stack level manually is error prone because
        # decorators like `@parameter_renamed` increase the call depth.)
        stacklevel = 0
        try:
            call_frame = introspection.CallFrame.up(1)
            stacklevel = 2

            while call_frame is not None:
                try:
                    module_name: str = call_frame.globals["__name__"]
                except KeyError:
                    # This can happen if the function is called via `eval` or
                    # `exec`.
                    break

                module_name, _, _ = module_name.partition(".")

                if (
                    module_name != "imy"
                    and module_name not in self.modules_skipped_in_stacktrace
                ):
                    break

                call_frame = call_frame.parent
                stacklevel += 1
        finally:
            call_frame = None  # Break the reference cycle

        warnings.warn(
            message,
            self.warning_class,
            stacklevel=stacklevel,
        )


DEFAULT_CONFIG = Config("imy")
module_name_to_config = dict[str, Config]()


def get_config_for_calling_module() -> Config:
    module_name = get_calling_module()
    return module_name_to_config.get(module_name, DEFAULT_CONFIG)


# I'm not a huge fan of inspecting the call stack, so this function exists as an
# alternatite to `get_config_for_calling_module()`. It obtains the module name
# via the `__module__` attribute of a class or function.
def get_config_for_defining_module(obj: t.Callable) -> Config:
    module_name, _, _ = obj.__module__.partition(".")
    return module_name_to_config.get(module_name, DEFAULT_CONFIG)


def configure(
    module: str | None = None,
    *,
    project_name: str | None = None,
    modules_skipped_in_stacktrace: t.Container[str] | None = None,
    warning_class: type[Warning] = DeprecationWarning,
    name_for_object: t.Callable[[t.Callable], str] = lambda obj: obj.__name__,
) -> None:
    """
    Configures how deprecation warnings are displayed for your module.

    ## Parameters

    `module`: The module to which this configuration should apply. This
        parameter is optional; if omitted, it will apply to the module that is
        calling this function.

    `project_name`: The name of your project, as it should apper in deprecation
        warnings.

    `modules_skipped_in_stacktrace`: Names of modules that should be omitted
        from the stacktrace of a deprecation warning.

    `warning_class`: The type of warning that is emitted at runtime.

        Note: Per default, python only displays `DeprecationWarning`s that
        originate from the `__main__` module. Keeping this in mind, you may want
        to choose a different `Warning` class or reconfigure the warnings
        filter.

    `name_for_object`: A function that takes a deprecated object as input and
        returns its name, as it should appear in deprecation warnings.
    """
    if module is None:
        module_name = get_calling_module()
    else:
        module_name, _, _ = module.partition(".")

    if project_name is None:
        project_name = module_name

    module_name_to_config[module_name] = Config(
        project_name=project_name,
        modules_skipped_in_stacktrace=modules_skipped_in_stacktrace,
        warning_class=warning_class,
        name_for_object=name_for_object,
    )


def get_calling_module() -> str:
    try:
        call_frame = introspection.CallFrame.up(1)

        while call_frame is not None:
            root_module, _, _ = call_frame.globals["__name__"].partition(".")

            if root_module != "imy":
                return root_module

            call_frame = call_frame.parent

        assert False, "All frames on the call stack are from `imy`?!"
    finally:
        call_frame = None  # Break the reference cycle


def function_or_class_decorator(
    decorator: t.Callable[[Config, C, t.Callable, t.Callable], C],
) -> t.Callable[[C], C]:
    """
    Many decorators in this module can be applied to both functions and classes,
    but classes need special handling - the decorator has to modify the class's
    `__init__` method instead.

    This function helps with the implementation of such decorators. It's hard to
    explain how it works, so here's an example instead:

    ```
    @function_or_class_decorator
    def my_decorator(
        config: Config,
        decorated_callable: C,
        undecorated_callable: t.Callable,
        function_to_wrap: t.Callable,
    ) -> C:
        ...
    ```

    The arguments passed to the decorator are:

    1. The relevant `Config`.
    2. The object that was decorated.
    3. The object that was decorated, but with all decorators removed. This is
       the "original" callable, for which the deprecation should be registered.
    4. The function that needs to be wrapped. (If the input was a class, this
       will be the relevant `__init__` method.)

    Your decorator doesn't need to use `@functools.wraps`. The metadata will be
    updated automatically.
    """

    def actual_decorator(decorated_callable: C) -> C:
        config = get_config_for_defining_module(decorated_callable)

        # Decorators often return a *new* callable, which is a problem for us.
        # We want to register all our Deprecations with the original,
        # undecorated, callable.
        undecorated_callable = inspect.unwrap(decorated_callable)

        # Easy case: It's a function. Just apply the decorator as usual.
        if not isinstance(decorated_callable, type):
            result = decorator(
                config,
                decorated_callable,
                undecorated_callable,
                decorated_callable,
            )
            functools.update_wrapper(result, decorated_callable)
            return result

        # Harder case: It's a class. We must decorate its `__init__` instead.
        original_init = decorated_callable.__init__
        new_init = decorator(
            config,
            decorated_callable,  # type: ignore (ugh)
            undecorated_callable,
            original_init,
        )

        new_init.__name__ = "__init__"
        new_init.__qualname__ = decorated_callable.__qualname__ + ".__init__"
        new_init.__module__ = decorated_callable.__module__
        new_init.__wrapped__ = original_init  # type: ignore

        decorated_callable.__init__ = new_init

        # Return the class, not the `__init__`
        return decorated_callable  # type: ignore (ugh)

    return actual_decorator


@t.overload
def deprecated(
    *,
    since: str,
    replacement: t.Callable | str,
    will_be_removed_in_version: str | None = None,
) -> t.Callable[[C], C]: ...


@t.overload
def deprecated(
    *,
    since: str,
    description: str,
    will_be_removed_in_version: str | None = None,
) -> t.Callable[[C], C]: ...


def deprecated(
    *,
    since: str,
    will_be_removed_in_version: str | None = None,
    description: str | None = None,
    replacement: t.Callable | str | None = None,
) -> t.Callable[[C], C]:
    """
    This is a class/function decorator that marks the decorated object as
    deprecated. Calling the deprecated object will emit a runtime warning.
    """

    @function_or_class_decorator
    def decorator(
        config: Config,
        decorated_callable: C,
        undecorated_callable: t.Callable,
        function_to_wrap: t.Callable,
    ) -> C:
        if description is None:
            warning_message = (
                f"`{config.name_for_object(decorated_callable)}` is deprecated"
            )
        else:
            warning_message = description

        if replacement is not None and not isinstance(replacement, str):
            replacement_name = config.name_for_object(replacement)
        else:
            replacement_name = replacement

        if replacement_name is not None:
            warning_message += f". Use `{replacement_name}` instead."

        deprecated_objects[undecorated_callable].append(
            Deprecation(
                since_version=since,
                message=warning_message,
                will_be_removed_in_version=will_be_removed_in_version,
            )
        )

        warning_message = config.add_versions(
            warning_message,
            since=since,
            will_be_removed_in_version=will_be_removed_in_version,
        )

        def wrapper(*args, **kwargs):
            config.warn(warning_message)
            return function_to_wrap(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def parameter_renamed(
    *,
    since: str,
    old_name: str,
    new_name: str,
    will_be_removed_in_version: str | None = None,
) -> t.Callable[[C], C]:
    """
    This decorator helps with renaming a parameter. Works on functions and
    classes.

    The old parameter should be removed from the signature. Example:

    ```
    @parameter_renamed(old_name='foo', new_name='bar', since='1.0')
    def my_function(bar: int):
        ...
    ```

    Note: In case of a class, only the `__init__` method will be modified.
    `__new__` methods are not affected.
    """

    @function_or_class_decorator
    def decorator(
        config: Config,
        decorated_callable: C,
        undecorated_callable: t.Callable,
        function_to_wrap: t.Callable,
    ) -> C:
        warning_message = config.make_parameter_deprecation_message(
            old_name=old_name,
            new_name=new_name,
            callable_=decorated_callable,
        )

        deprecated_parameters[undecorated_callable][old_name].append(
            Deprecation(
                since_version=since,
                message=warning_message,
                will_be_removed_in_version=will_be_removed_in_version,
            )
        )

        warning_message = config.add_versions(
            warning_message,
            since=since,
            will_be_removed_in_version=will_be_removed_in_version,
        )

        def new_function(*args, **kwargs):
            # Remap the old parameter to the new one
            try:
                kwargs[new_name] = kwargs.pop(old_name)
            except KeyError:
                pass
            else:
                config.warn(warning_message)

            # Delegate to the original function
            return function_to_wrap(*args, **kwargs)

        return new_function  # type: ignore

    return decorator


def parameter_remapped(
    *,
    since: str,
    old_name: str,
    new_name: str,
    remap: t.Callable[[t.Any], t.Any],
    will_be_removed_in_version: str | None = None,
):
    """
    This is a function decorator that's quite similar to `parameter_renamed`,
    but it allows you to change the type and value of the parameter as well
    as the name.

    Example: `Theme.from_colors` used to have a `light: bool = True` parameter
    which was changed to `mode: t.Literal['light', 'dark'] = 'light'`.

        class Theme:
            @parameter_remapped(
                since='0.9',
                old_name='light',
                new_name='mode'
                remap=lambda light: "light" if light else "dark",
            )
            def from_colors(..., mode: t.Literal['light', 'dark'] = 'light'):
                ...

        Theme.from_colors(light=False)  # Equivalent to `mode='dark'`

    WARNING: The remapping only takes effect if the argument is passed as a
    keyword argument.
    """

    @function_or_class_decorator
    def decorator(
        config: Config,
        decorated_callable: C,
        undecorated_callable: t.Callable,
        function_to_wrap: t.Callable,
    ) -> C:
        warning_message = config.make_parameter_deprecation_message(
            old_name=old_name,
            new_name=new_name,
            callable_=decorated_callable,
        )

        deprecated_parameters[undecorated_callable][old_name].append(
            Deprecation(
                since_version=since,
                message=warning_message,
                will_be_removed_in_version=will_be_removed_in_version,
            )
        )

        warning_message = config.add_versions(
            warning_message,
            since=since,
            will_be_removed_in_version=will_be_removed_in_version,
        )

        def wrapper(*args, **kwargs):
            try:
                old_value = kwargs.pop(old_name)
            except KeyError:
                pass
            else:
                config.warn(warning_message)
                kwargs[new_name] = remap(old_value)

            return function_to_wrap(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def warn(message: str, *, since: str | None = None):
    """
    Emits a runtime warning.

    This function is part of the "low-level API". Please use the provided
    decorators instead whenever possible.
    """
    config = get_config_for_calling_module()

    if since is not None:
        message = config.add_versions(message, since=since)

    config.warn(message)


def warn_parameter_renamed(
    *,
    old_name: str,
    new_name: str,
    function: t.Callable | str | None = None,
    since: str | None = None,
    will_be_removed_in_version: str | None = None,
):
    """
    Emits a runtime warning about a parameter being renamed.

    This function is part of the "low-level API". Please use the provided
    decorators instead whenever possible.
    """
    config = get_config_for_calling_module()

    message = config.make_parameter_deprecation_message(
        old_name=old_name, new_name=new_name, callable_=function
    )

    message = config.add_versions(
        message,
        since=since,
        will_be_removed_in_version=will_be_removed_in_version,
    )

    config.warn(message)



"""
Functionality for standardized docstring parsing.

Docstrings are expected to have the following form:

```
# An optional heading-1 in the first line is stripped

The first line is the summary.

The summary is followed by a blank line and then the description. The
description may be arbitrarily long. It may span multiple lines, paragraphs,
contain code blocks, whatever.

## Parameters

`param_`: Parameters are stored in key sections. Values may span multiple
    lines, as long as followup lines are indented

param_2: The backticks around the name are optional, and only so the name shows
    up as code when viewing the docstring directly inside a code editor

## Raises

`ValueError`: This section works the same as parameters

## Attributes

`attr_1`: As does this

## Metadata

`public`: Whether this object is meant for anyone outside of the packages itself

`experimental`: This marks code whose API may still change

## Other Sections

Any sections other than the ones listed above are considered to be part of the
description. They are not parsed in any way.
```
"""

from .data_models import *
from .utilities import *



import pytest

import imy.inject


class A:
    pass


class B:
    pass


class B_A(B):
    pass


class C:
    pass


class D:
    pass


class E:
    pass


def make_injector() -> imy.inject.Injector:
    """
    Creates a fresh injector without any instantiated components.
    """
    injector = imy.inject.Injector()

    @injector.bind()
    def create_a() -> A:
        return A()

    @injector.bind(B)
    def create_b_a() -> B_A:
        return B_A()

    @injector.bind()
    def create_c(a: A, b: B) -> C:
        assert isinstance(a, A)
        assert isinstance(b, B)
        return C()

    @injector.bind()
    def create_d(e: E) -> D:
        assert isinstance(e, E)
        return D()

    @injector.bind()
    def create_e(d: D) -> E:
        assert isinstance(d, D)
        return E()

    return injector


def test_get_a() -> None:
    injector = make_injector()
    a = injector[A]
    assert isinstance(a, A)


def test_get_b() -> None:
    injector = make_injector()
    b = injector[B]
    assert isinstance(b, B_A)


def test_get_c() -> None:
    injector = make_injector()
    c = injector[C]
    assert isinstance(c, C)


def test_get_d() -> None:
    injector = make_injector()

    with pytest.raises(imy.inject.DependencyCycleError):
        injector[D]


def test_inject_subtyped() -> None:
    """
    Injectors aren't supposed to distinguish by the "args" of a type, i.e.
    `list[int]` and `list[str]` are the same type to them. Verify that.
    """
    injector = imy.inject.Injector()

    @injector.bind()
    def create_list_int() -> list[int]:
        return [1, 2, 3]

    result = injector[list]
    assert result == [1, 2, 3]


def test_inject_subtyped_clash() -> None:
    """
    Make sure injectors don't allow two different subtypes of the same type.
    """
    injector = imy.inject.Injector()

    @injector.bind()
    def create_list_int() -> list[int]:
        return [1, 2, 3]

    with pytest.raises(TypeError):

        @injector.bind()
        def create_list_str() -> list[str]:
            return ["a", "b", "c"]


def test_duplicate_loggers_with_clear() -> None:
    """
    Registering duplicate factories is fine, as long as the injector is cleared
    before the second registration.
    """

    injector = imy.inject.Injector()

    @injector.bind()
    def create_a() -> A:
        return A()

    injector.clear()

    @injector.bind()
    def create_a_again() -> A:
        return A()
